@@ -18,7 +18,7 @@ References
 
 from __future__ import annotations
 
-from typing import Any, Callable, List, Optional, Union
+from collections.abc import Callable
 
 import torch
 from botorch.acquisition.cost_aware import InverseCostWeightedUtility
@@ -33,7 +33,11 @@ from botorch.sampling.base import MCSampler
 from botorch.utils.multi_objective.box_decompositions.non_dominated import (
     NondominatedPartitioning,
 )
-from botorch.utils.transforms import concatenate_pending_points, t_batch_mode_transform
+from botorch.utils.transforms import (
+    average_over_ensemble_models,
+    concatenate_pending_points,
+    t_batch_mode_transform,
+)
 from torch import Tensor
 
 
@@ -41,15 +45,14 @@ class MOMF(qExpectedHypervolumeImprovement):
     def __init__(
         self,
         model: Model,
-        ref_point: Union[List[float], Tensor],
+        ref_point: list[float] | Tensor,
         partitioning: NondominatedPartitioning,
-        sampler: Optional[MCSampler] = None,
-        objective: Optional[MCMultiOutputObjective] = None,
-        constraints: Optional[List[Callable[[Tensor], Tensor]]] = None,
-        eta: Optional[Union[Tensor, float]] = 1e-3,
-        X_pending: Optional[Tensor] = None,
-        cost_call: Callable[Tensor, Tensor] = None,
-        **kwargs: Any,
+        sampler: MCSampler | None = None,
+        objective: MCMultiOutputObjective | None = None,
+        constraints: list[Callable[[Tensor], Tensor]] | None = None,
+        eta: Tensor | float = 1e-3,
+        X_pending: Tensor | None = None,
+        cost_call: Callable[[Tensor], Tensor] | None = None,
     ) -> None:
         r"""MOMF acquisition function supporting m>=2 outcomes.
         The model needs to have train_obj that has a fidelity
@@ -83,7 +86,7 @@ class MOMF(qExpectedHypervolumeImprovement):
             sampler: The sampler used to draw base samples. If not given,
                 a sampler is generated using `get_sampler`.
             objective: The MCMultiOutputObjective under which the samples are evaluated.
-                Defaults to `IdentityMultiOutputObjective()`.
+                Defaults to `IdentityMCMultiOutputObjective()`.
             constraints: A list of callables, each mapping a Tensor of dimension
                 `sample_shape x batch-shape x q x m` to a Tensor of dimension
                 `sample_shape x batch-shape x q`, where negative values imply
@@ -138,6 +141,7 @@ class MOMF(qExpectedHypervolumeImprovement):
 
     @concatenate_pending_points
     @t_batch_mode_transform()
+    @average_over_ensemble_models
     def forward(self, X: Tensor) -> Tensor:
         posterior = self.model.posterior(X)
         samples = self.get_posterior_samples(posterior)
