@@ -11,7 +11,8 @@ Modules to add regularization to acquisition functions.
 from __future__ import annotations
 
 import math
-from typing import Callable, List, Optional
+from collections.abc import Callable
+from typing import Any
 
 import torch
 from botorch.acquisition.acquisition import AcquisitionFunction
@@ -44,7 +45,8 @@ class L2Penalty(torch.nn.Module):
             A tensor of size "batch_shape" representing the acqfn for each q-batch.
         """
         regularization_term = (
-            torch.norm((X - self.init_point), p=2, dim=-1).max(dim=-1).values ** 2
+            torch.linalg.norm((X - self.init_point), ord=2, dim=-1).max(dim=-1).values
+            ** 2
         )
         return regularization_term
 
@@ -72,7 +74,7 @@ class L1Penalty(torch.nn.Module):
             A tensor of size "batch_shape" representing the acqfn for each q-batch.
         """
         regularization_term = (
-            torch.norm((X - self.init_point), p=1, dim=-1).max(dim=-1).values
+            torch.linalg.norm((X - self.init_point), ord=1, dim=-1).max(dim=-1).values
         )
         return regularization_term
 
@@ -101,7 +103,7 @@ class GaussianPenalty(torch.nn.Module):
         Returns:
             A tensor of size "batch_shape" representing the acqfn for each q-batch.
         """
-        sq_diff = torch.norm((X - self.init_point), p=2, dim=-1) ** 2
+        sq_diff = torch.linalg.norm((X - self.init_point), ord=2, dim=-1) ** 2
         pdf = torch.exp(sq_diff / 2 / self.sigma**2)
         regularization_term = pdf.max(dim=-1).values
         return regularization_term
@@ -111,7 +113,7 @@ class GroupLassoPenalty(torch.nn.Module):
     r"""Group lasso penalty class to be added to any arbitrary acquisition function
     to construct a PenalizedAcquisitionFunction."""
 
-    def __init__(self, init_point: Tensor, groups: List[List[int]]):
+    def __init__(self, init_point: Tensor, groups: list[list[int]]):
         r"""Initializing Group-Lasso regularization.
 
         Args:
@@ -139,6 +141,66 @@ class GroupLassoPenalty(torch.nn.Module):
         return regularization_term
 
 
+def narrow_gaussian(X: Tensor, a: Tensor) -> Tensor:
+    return torch.exp(-0.5 * (X / a) ** 2)
+
+
+def nnz_approx(X: Tensor, target_point: Tensor, a: Tensor) -> Tensor:
+    r"""Differentiable relaxation of ||X - target_point||_0
+
+    Args:
+        X: An `n x d` tensor of inputs.
+        target_point: A tensor of size `n` corresponding to the target point.
+        a: A scalar tensor that controls the differentiable relaxation.
+    """
+    d = X.shape[-1]
+    if d != target_point.shape[-1]:
+        raise ValueError("X and target_point have different shapes.")
+    return d - narrow_gaussian(X - target_point, a).sum(dim=-1, keepdim=True)
+
+
+class L0Approximation(torch.nn.Module):
+    r"""Differentiable relaxation of the L0 norm using a Gaussian basis function."""
+
+    def __init__(self, target_point: Tensor, a: float = 1.0, **tkwargs: Any) -> None:
+        r"""Initializing L0 penalty with differentiable relaxation.
+
+        Args:
+            target_point: A tensor corresponding to the target point.
+            a: A hyperparameter that controls the differentiable relaxation.
+        """
+        super().__init__()
+        self.target_point = target_point
+        # hyperparameter to control the differentiable relaxation in L0 norm function.
+        self.register_buffer("a", torch.tensor(a, **tkwargs))
+
+    def __call__(self, X: Tensor) -> Tensor:
+        return nnz_approx(X=X, target_point=self.target_point, a=self.a)
+
+
+class L0PenaltyApprox(L0Approximation):
+    r"""Differentiable relaxation of the L0 norm to be added to any arbitrary
+    acquisition function to construct a PenalizedAcquisitionFunction."""
+
+    def __init__(self, target_point: Tensor, a: float = 1.0, **tkwargs: Any) -> None:
+        r"""Initializing L0 penalty with differentiable relaxation.
+
+        Args:
+            target_point: A tensor corresponding to the target point.
+            a: A hyperparameter that controls the differentiable relaxation.
+        """
+        super().__init__(target_point=target_point, a=a, **tkwargs)
+
+    def __call__(self, X: Tensor) -> Tensor:
+        r"""
+        Args:
+            X: A "batch_shape x q x dim" representing the points to be evaluated.
+        Returns:
+            A tensor of size "batch_shape" representing the acqfn for each q-batch.
+        """
+        return super().__call__(X=X).squeeze(dim=-1).min(dim=-1).values
+
+
 class PenalizedAcquisitionFunction(AcquisitionFunction):
     r"""Single-outcome acquisition function regularized by the given penalty.
 
@@ -172,10 +234,10 @@ class PenalizedAcquisitionFunction(AcquisitionFunction):
         return raw_value - self.regularization_parameter * penalty_term
 
     @property
-    def X_pending(self) -> Optional[Tensor]:
+    def X_pending(self) -> Tensor | None:
         return self.raw_acqf.X_pending
 
-    def set_X_pending(self, X_pending: Optional[Tensor] = None) -> None:
+    def set_X_pending(self, X_pending: Tensor | None = None) -> None:
         if not isinstance(self.raw_acqf, AnalyticAcquisitionFunction):
             self.raw_acqf.set_X_pending(X_pending=X_pending)
         else:
@@ -185,7 +247,7 @@ class PenalizedAcquisitionFunction(AcquisitionFunction):
             )
 
 
-def group_lasso_regularizer(X: Tensor, groups: List[List[int]]) -> Tensor:
+def group_lasso_regularizer(X: Tensor, groups: list[list[int]]) -> Tensor:
     r"""Computes the group lasso regularization function for the given point.
 
     Args:
@@ -197,7 +259,10 @@ def group_lasso_regularizer(X: Tensor, groups: List[List[int]]) -> Tensor:
     """
     return torch.sum(
         torch.stack(
-            [math.sqrt(len(g)) * torch.norm(X[..., g], p=2, dim=-1) for g in groups],
+            [
+                math.sqrt(len(g)) * torch.linalg.norm(X[..., g], ord=2, dim=-1)
+                for g in groups
+            ],
             dim=-1,
         ),
         dim=-1,
@@ -229,13 +294,13 @@ class L1PenaltyObjective(torch.nn.Module):
             A "1 x batch_shape x q" tensor representing the penalty for each point.
             The first dimension corresponds to the dimension of MC samples.
         """
-        return torch.norm((X - self.init_point), p=1, dim=-1).unsqueeze(dim=0)
+        return torch.linalg.norm((X - self.init_point), ord=1, dim=-1).unsqueeze(dim=0)
 
 
 class PenalizedMCObjective(GenericMCObjective):
     r"""Penalized MC objective.
 
-    Allows to construct a penaltized MC-objective by adding a penalty term to
+    Allows to construct a penalized MC-objective by adding a penalty term to
     the original objective.
 
         mc_acq(X) = objective(X) + penalty_objective(X)
@@ -257,9 +322,10 @@ class PenalizedMCObjective(GenericMCObjective):
 
     def __init__(
         self,
-        objective: Callable[[Tensor, Optional[Tensor]], Tensor],
+        objective: Callable[[Tensor, Tensor | None], Tensor],
         penalty_objective: torch.nn.Module,
         regularization_parameter: float,
+        expand_dim: int | None = None,
     ) -> None:
         r"""Penalized MC objective.
 
@@ -272,12 +338,15 @@ class PenalizedMCObjective(GenericMCObjective):
                 `batch-shape x q x d`-dim Tensor `X` and outputs a
                 `1 x batch-shape x q`-dim Tensor of penalty objective values.
             regularization_parameter: weight of the penalty (regularization) term
+            expand_dim: dim to expand penalty_objective to match with objective when
+                fully bayesian model is used. If None, no expansion is performed.
         """
         super().__init__(objective=objective)
         self.penalty_objective = penalty_objective
         self.regularization_parameter = regularization_parameter
+        self.expand_dim = expand_dim
 
-    def forward(self, samples: Tensor, X: Optional[Tensor] = None) -> Tensor:
+    def forward(self, samples: Tensor, X: Tensor | None = None) -> Tensor:
         r"""Evaluate the penalized objective on the samples.
 
         Args:
@@ -292,4 +361,42 @@ class PenalizedMCObjective(GenericMCObjective):
         """
         obj = super().forward(samples=samples, X=X)
         penalty_obj = self.penalty_objective(X)
+        # when fully bayesian model is used, we pass unmarginalize_dim to match the
+        # shape between obj `sample_shape x batch-shape x mcmc_samples x q` and
+        # penalty_obj `1 x batch-shape x q`
+        if self.expand_dim is not None:
+            # reshape penalty_obj to match the dim
+            penalty_obj = penalty_obj.unsqueeze(self.expand_dim)
+        # this happens when samples is a `q x m`-dim tensor and X is a `q x d`-dim
+        # tensor; obj returned from GenericMCObjective is a `q`-dim tensor and
+        # penalty_obj is a `1 x q`-dim tensor.
+        if obj.ndim == 1:
+            assert penalty_obj.shape == torch.Size([1, samples.shape[-2]])
+            penalty_obj = penalty_obj.squeeze(dim=0)
         return obj - self.regularization_parameter * penalty_obj
+
+
+class L0PenaltyApproxObjective(L0Approximation):
+    r"""Differentiable relaxation of the L0 norm penalty objective class.
+    An instance of this class can be added to any arbitrary objective to
+    construct a PenalizedMCObjective.
+    """
+
+    def __init__(self, target_point: Tensor, a: float = 1.0, **tkwargs: Any) -> None:
+        r"""Initializing L0 penalty with differentiable relaxation.
+
+        Args:
+            target_point: A tensor corresponding to the target point.
+            a: A hyperparameter that controls the differentiable relaxation.
+        """
+        super().__init__(target_point=target_point, a=a, **tkwargs)
+
+    def __call__(self, X: Tensor) -> Tensor:
+        r"""
+        Args:
+            X: A "batch_shape x q x dim" representing the points to be evaluated.
+        Returns:
+            A "1 x batch_shape x q" tensor representing the penalty for each point.
+            The first dimension corresponds to the dimension of MC samples.
+        """
+        return super().__call__(X=X).squeeze(dim=-1).unsqueeze(dim=0)
