@@ -6,22 +6,23 @@
 
 from __future__ import annotations
 
-from typing import Any, Callable, Dict, List, Optional
+from collections.abc import Callable
+
+from typing import Any
 
 import torch
 from botorch.models.gp_regression import SingleTaskGP
 from botorch.models.kernels.categorical import CategoricalKernel
 from botorch.models.transforms.input import InputTransform
 from botorch.models.transforms.outcome import OutcomeTransform
+from botorch.models.utils.gpytorch_modules import get_covar_module_with_dim_scaled_prior
 from botorch.utils.datasets import SupervisedDataset
 from botorch.utils.transforms import normalize_indices
+from botorch.utils.types import _DefaultType, DEFAULT
 from gpytorch.constraints import GreaterThan
 from gpytorch.kernels.kernel import Kernel
-from gpytorch.kernels.matern_kernel import MaternKernel
 from gpytorch.kernels.scale_kernel import ScaleKernel
-from gpytorch.likelihoods.gaussian_likelihood import GaussianLikelihood
 from gpytorch.likelihoods.likelihood import Likelihood
-from gpytorch.priors import GammaPrior
 from torch import Tensor
 
 
@@ -61,11 +62,13 @@ class MixedSingleTaskGP(SingleTaskGP):
         self,
         train_X: Tensor,
         train_Y: Tensor,
-        cat_dims: List[int],
-        cont_kernel_factory: Optional[Callable[[int, List[int]], Kernel]] = None,
-        likelihood: Optional[Likelihood] = None,
-        outcome_transform: Optional[OutcomeTransform] = None,  # TODO
-        input_transform: Optional[InputTransform] = None,  # TODO
+        cat_dims: list[int],
+        train_Yvar: Tensor | None = None,
+        cont_kernel_factory: None
+        | (Callable[[torch.Size, int, list[int]], Kernel]) = None,
+        likelihood: Likelihood | None = None,
+        outcome_transform: OutcomeTransform | _DefaultType | None = DEFAULT,
+        input_transform: InputTransform | None = None,  # TODO
     ) -> None:
         r"""A single-task exact GP model supporting categorical parameters.
 
@@ -74,68 +77,37 @@ class MixedSingleTaskGP(SingleTaskGP):
             train_Y: A `batch_shape x n x m` tensor of training observations.
             cat_dims: A list of indices corresponding to the columns of
                 the input `X` that should be considered categorical features.
-            cont_kernel_factory: A method that accepts `ard_num_dims` and
-                `active_dims` arguments and returns an instantiated GPyTorch
-                `Kernel` object to be used as the ase kernel for the continuous
-                dimensions. If omitted, this model uses a Matern-2.5 kernel as
+            train_Yvar: An optional `batch_shape x n x m` tensor of observed
+                measurement noise.
+            cont_kernel_factory: A method that accepts  `batch_shape`, `ard_num_dims`,
+                and `active_dims` arguments and returns an instantiated GPyTorch
+                `Kernel` object to be used as the base kernel for the continuous
+                dimensions. If omitted, this model uses an `RBFKernel` as
                 the kernel for the ordinal parameters.
             likelihood: A likelihood. If omitted, use a standard
                 GaussianLikelihood with inferred noise level.
             outcome_transform: An outcome transform that is applied to the
                 training data during instantiation and to the posterior during
                 inference (that is, the `Posterior` obtained by calling
-                `.posterior` on the model will be on the original scale).
+                `.posterior` on the model will be on the original scale). We use a
+                `Standardize` transform if no `outcome_transform` is specified.
+                Pass down `None` to use no outcome transform.
             input_transform: An input transform that is applied in the model's
                 forward pass. Only input transforms are allowed which do not
-                transform the categorical dimensions. This can be achieved
-                by using the `indices` argument when constructing the transform.
+                transform the categorical dimensions. If you want to use it
+                for example in combination with a `OneHotToNumeric` input transform
+                one has to instantiate the transform with `transform_on_train` == False
+                and pass in the already transformed input.
         """
-        if input_transform is not None:
-            if not hasattr(input_transform, "indices"):
-                raise ValueError(
-                    "Only continuous inputs can be transformed. "
-                    "Please use `indices` in the `input_transform`."
-                )
-            # check that no cat dim is in indices
-            elif any(idx in input_transform.indices for idx in cat_dims):
-                raise ValueError(
-                    "Only continuous inputs can be transformed. "
-                    "Categorical index found in `indices` of the `input_transform`."
-                )
         if len(cat_dims) == 0:
             raise ValueError(
                 "Must specify categorical dimensions for MixedSingleTaskGP"
             )
         self._ignore_X_dims_scaling_check = cat_dims
-        input_batch_shape, aug_batch_shape = self.get_batch_dimensions(
-            train_X=train_X, train_Y=train_Y
-        )
+        _, aug_batch_shape = self.get_batch_dimensions(train_X=train_X, train_Y=train_Y)
 
         if cont_kernel_factory is None:
-
-            def cont_kernel_factory(
-                batch_shape: torch.Size,
-                ard_num_dims: int,
-                active_dims: List[int],
-            ) -> MaternKernel:
-                return MaternKernel(
-                    nu=2.5,
-                    batch_shape=batch_shape,
-                    ard_num_dims=ard_num_dims,
-                    active_dims=active_dims,
-                    lengthscale_constraint=GreaterThan(1e-04),
-                )
-
-        if likelihood is None:
-            # This Gamma prior is quite close to the Horseshoe prior
-            min_noise = 1e-5 if train_X.dtype == torch.float else 1e-6
-            likelihood = GaussianLikelihood(
-                batch_shape=aug_batch_shape,
-                noise_constraint=GreaterThan(
-                    min_noise, transform=None, initial_value=1e-3
-                ),
-                noise_prior=GammaPrior(0.9, 10.0),
-            )
+            cont_kernel_factory = get_covar_module_with_dim_scaled_prior
 
         d = train_X.shape[-1]
         cat_dims = normalize_indices(indices=cat_dims, d=d)
@@ -181,6 +153,7 @@ class MixedSingleTaskGP(SingleTaskGP):
         super().__init__(
             train_X=train_X,
             train_Y=train_Y,
+            train_Yvar=train_Yvar,
             likelihood=likelihood,
             covar_module=covar_module,
             outcome_transform=outcome_transform,
@@ -191,19 +164,19 @@ class MixedSingleTaskGP(SingleTaskGP):
     def construct_inputs(
         cls,
         training_data: SupervisedDataset,
-        categorical_features: List[int],
-        likelihood: Optional[Likelihood] = None,
-        **kwargs: Any,
-    ) -> Dict[str, Any]:
-        r"""Construct `Model` keyword arguments from a dict of `BotorchDataset`.
+        categorical_features: list[int],
+        likelihood: Likelihood | None = None,
+    ) -> dict[str, Any]:
+        r"""Construct `Model` keyword arguments from a dict of `SupervisedDataset`.
 
         Args:
             training_data: A `SupervisedDataset` containing the training data.
             categorical_features: Column indices of categorical features.
             likelihood: Optional likelihood used to constuct the model.
         """
+        base_inputs = super().construct_inputs(training_data=training_data)
         return {
-            **super().construct_inputs(training_data=training_data, **kwargs),
+            **base_inputs,
             "cat_dims": categorical_features,
             "likelihood": likelihood,
         }
