@@ -15,37 +15,33 @@ from __future__ import annotations
 import warnings
 from abc import ABC, abstractmethod
 from collections import defaultdict
-from copy import deepcopy
-from typing import (
-    Any,
-    Callable,
-    Dict,
-    Hashable,
-    List,
-    Mapping,
-    Optional,
-    TYPE_CHECKING,
-    TypeVar,
-    Union,
-)
+from collections.abc import Callable, Mapping
+from typing import Any, TYPE_CHECKING
 
 import numpy as np
 import torch
 from botorch import settings
-from botorch.exceptions.errors import InputDataError
+from botorch.exceptions.errors import (
+    BotorchTensorDimensionError,
+    DeprecationError,
+    InputDataError,
+)
+from botorch.logging import shape_to_str
 from botorch.models.utils.assorted import fantasize as fantasize_flag
 from botorch.posteriors import Posterior, PosteriorList
 from botorch.sampling.base import MCSampler
-from botorch.utils.datasets import BotorchDataset
+from botorch.sampling.list_sampler import ListSampler
+from botorch.utils.containers import BotorchContainer
+from botorch.utils.datasets import SupervisedDataset
 from botorch.utils.transforms import is_fully_bayesian
+from gpytorch.likelihoods.gaussian_likelihood import FixedNoiseGaussianLikelihood
 from torch import Tensor
 from torch.nn import Module, ModuleDict, ModuleList
+from typing_extensions import Self
 
 if TYPE_CHECKING:
     from botorch.acquisition.objective import PosteriorTransform  # pragma: no cover
 
-TFantasizeMixin = TypeVar("TFantasizeMixin", bound="FantasizeMixin")
-
 
 class Model(Module, ABC):
     r"""Abstract base class for BoTorch models.
@@ -63,26 +59,31 @@ class Model(Module, ABC):
     `Tensor` or `Module` type are automatically registered so they can be moved and/or
     cast with the `to` method, automatically differentiated, and used with CUDA.
 
-    Args:
+    Attributes:
         _has_transformed_inputs: A boolean denoting whether `train_inputs` are currently
             stored as transformed or not.
         _original_train_inputs: A Tensor storing the original train inputs for use in
             `_revert_to_original_inputs`. Note that this is necessary since
             transform / untransform cycle introduces numerical errors which lead
             to upstream errors during training.
+        _is_fully_bayesian: Returns `True` if this is a fully Bayesian model.
+        _is_ensemble: Returns `True` if this model consists of multiple models
+            that are stored in an additional batch dimension. This is true for the fully
+            Bayesian models.
     """  # noqa: E501
 
     _has_transformed_inputs: bool = False
-    _original_train_inputs: Optional[Tensor] = None
+    _original_train_inputs: Tensor | None = None
+    _is_fully_bayesian = False
+    _is_ensemble = False
 
     @abstractmethod
     def posterior(
         self,
         X: Tensor,
-        output_indices: Optional[List[int]] = None,
-        observation_noise: bool = False,
-        posterior_transform: Optional[PosteriorTransform] = None,
-        **kwargs: Any,
+        output_indices: list[int] | None = None,
+        observation_noise: bool | Tensor = False,
+        posterior_transform: PosteriorTransform | None = None,
     ) -> Posterior:
         r"""Computes the posterior over model outputs at the provided points.
 
@@ -99,7 +100,12 @@ class Model(Module, ABC):
                 Can be used to speed up computation if only a subset of the
                 model's outputs are required for optimization. If omitted,
                 computes the posterior over all model outputs.
-            observation_noise: If True, add observation noise to the posterior.
+            observation_noise: For models with an inferred noise level, if True,
+                include observation noise. For models with an observed noise level,
+                this must be a `model_batch_shape x 1 x m`-dim tensor or
+                a `model_batch_shape x n' x m`-dim tensor containing the average
+                noise for each batch and output. `noise` must be in the
+                outcome-transformed space if an outcome transform is used.
             posterior_transform: An optional PosteriorTransform.
 
         Returns:
@@ -127,7 +133,7 @@ class Model(Module, ABC):
         cls_name = self.__class__.__name__
         raise NotImplementedError(f"{cls_name} does not define num_outputs property")
 
-    def subset_output(self, idcs: List[int]) -> Model:
+    def subset_output(self, idcs: list[int]) -> Model:
         r"""Subset the model along the output dimension.
 
         Args:
@@ -167,18 +173,33 @@ class Model(Module, ABC):
     @classmethod
     def construct_inputs(
         cls,
-        training_data: Union[BotorchDataset, Dict[Hashable, BotorchDataset]],
-        **kwargs: Any,
-    ) -> Dict[str, Any]:
-        r"""Construct `Model` keyword arguments from a dict of `BotorchDataset`."""
-        from botorch.models.utils.parse_training_data import parse_training_data
+        training_data: SupervisedDataset,
+    ) -> dict[str, BotorchContainer | Tensor]:
+        """
+        Construct `Model` keyword arguments from a `SupervisedDataset`.
 
-        return parse_training_data(cls, training_data, **kwargs)
+        Args:
+            training_data: A `SupervisedDataset`, with attributes `train_X`,
+                `train_Y`, and, optionally, `train_Yvar`.
+
+        Returns:
+            A dict of keyword arguments that can be used to initialize a `Model`,
+            with keys `train_X`, `train_Y`, and, optionally, `train_Yvar`.
+        """
+        if not isinstance(training_data, SupervisedDataset):
+            raise TypeError(
+                "Expected `training_data` to be a `SupervisedDataset`, but got "
+                f"{type(training_data)}."
+            )
+        parsed_data = {"train_X": training_data.X, "train_Y": training_data.Y}
+        if training_data.Yvar is not None:
+            parsed_data["train_Yvar"] = training_data.Yvar
+        return parsed_data
 
     def transform_inputs(
         self,
         X: Tensor,
-        input_transform: Optional[Module] = None,
+        input_transform: Module | None = None,
     ) -> Tensor:
         r"""Transform inputs.
 
@@ -215,6 +236,7 @@ class Model(Module, ABC):
                     "attribute. Make sure that the `input_transform` is applied to "
                     "both the train inputs and test inputs.",
                     RuntimeWarning,
+                    stacklevel=3,
                 )
 
     def _revert_to_original_inputs(self) -> None:
@@ -242,6 +264,10 @@ class Model(Module, ABC):
             self._set_transformed_inputs()
         return super().train(mode=mode)
 
+    @property
+    def dtypes_of_buffers(self) -> set[torch.dtype]:
+        return {t.dtype for t in self.buffers() if t is not None}
+
 
 class FantasizeMixin(ABC):
     """
@@ -263,9 +289,7 @@ class FantasizeMixin(ABC):
     """
 
     @abstractmethod
-    def condition_on_observations(
-        self: TFantasizeMixin, X: Tensor, Y: Tensor, **kwargs: Any
-    ) -> TFantasizeMixin:
+    def condition_on_observations(self, X: Tensor, Y: Tensor) -> Self:
         """
         Classes that inherit from `FantasizeMixin` must implement
         a `condition_on_observations` method.
@@ -277,7 +301,6 @@ class FantasizeMixin(ABC):
         X: Tensor,
         *args,
         observation_noise: bool = False,
-        **kwargs: Any,
     ) -> Posterior:
         """
         Classes that inherit from `FantasizeMixin` must implement
@@ -288,29 +311,26 @@ class FantasizeMixin(ABC):
     def transform_inputs(
         self,
         X: Tensor,
-        input_transform: Optional[Module] = None,
+        input_transform: Module | None = None,
     ) -> Tensor:
         """
         Classes that inherit from `FantasizeMixin` must implement
         a `transform_inputs` method.
         """
 
-    # When Python 3.11 arrives we can start annotating return types like
-    # this as
-    # 'Self', but at this point the verbose 'T...' syntax is needed.
     def fantasize(
-        self: TFantasizeMixin,
-        # TODO: see if any of these can be imported only if TYPE_CHECKING
+        self,
         X: Tensor,
         sampler: MCSampler,
-        observation_noise: bool = True,
+        observation_noise: Tensor | None = None,
         **kwargs: Any,
-    ) -> TFantasizeMixin:
+    ) -> Self:
         r"""Construct a fantasy model.
 
         Constructs a fantasy model in the following fashion:
-        (1) compute the model posterior at `X` (including observation noise if
-        `observation_noise=True`).
+        (1) compute the model posterior at `X`, including observation noise.
+        If `observation_noise` is a Tensor, use it directly as the observation
+        noise to add.
         (2) sample from this posterior (using `sampler`) to generate "fake"
         observations.
         (3) condition the model on the new fake observations.
@@ -321,17 +341,60 @@ class FantasizeMixin(ABC):
                 `batch_shape` is the batch shape (must be compatible with the
                 batch shape of the model).
             sampler: The sampler used for sampling from the posterior at `X`.
-            observation_noise: If True, include observation noise.
+            observation_noise: A `model_batch_shape x 1 x m`-dim tensor or
+                a `model_batch_shape x n' x m`-dim tensor containing the average
+                noise for each batch and output, where `m` is the number of outputs.
+                `noise` must be in the outcome-transformed space if an outcome
+                transform is used.
+                If None and using an inferred noise likelihood, the noise will be the
+                inferred noise level. If using a fixed noise likelihood, the mean across
+                the observation noise in the training data is used as observation noise.
             kwargs: Will be passed to `model.condition_on_observations`
 
         Returns:
             The constructed fantasy model.
         """
+        if not isinstance(observation_noise, Tensor) and observation_noise is not None:
+            raise DeprecationError(
+                "`fantasize` no longer accepts a boolean for `observation_noise`."
+            )
+        elif observation_noise is None and isinstance(
+            self.likelihood, FixedNoiseGaussianLikelihood
+        ):
+            if self.num_outputs > 1:
+                # make noise ... x n x m
+                observation_noise = self.likelihood.noise.transpose(-1, -2)
+            else:
+                observation_noise = self.likelihood.noise.unsqueeze(-1)
+            observation_noise = observation_noise.mean(dim=-2, keepdim=True)
+        # if the inputs are empty, expand the inputs
+        if X.shape[-2] == 0:
+            output_shape = (
+                sampler.sample_shape
+                + X.shape[:-2]
+                + self.batch_shape
+                + torch.Size([0, self.num_outputs])
+            )
+            Y = torch.empty(output_shape, dtype=X.dtype, device=X.device)
+            if observation_noise is not None:
+                kwargs["noise"] = observation_noise.expand(Y.shape[1:])
+            return self.condition_on_observations(
+                X=self.transform_inputs(X),
+                Y=Y,
+                **kwargs,
+            )
         propagate_grads = kwargs.pop("propagate_grads", False)
         with fantasize_flag():
             with settings.propagate_grads(propagate_grads):
-                post_X = self.posterior(X, observation_noise=observation_noise)
+                post_X = self.posterior(
+                    X,
+                    observation_noise=(
+                        True if observation_noise is None else observation_noise
+                    ),
+                )
             Y_fantasized = sampler(post_X)  # num_fantasies x batch_shape x n' x m
+            if observation_noise is not None:
+                kwargs["noise"] = observation_noise.expand(Y_fantasized.shape[1:])
             return self.condition_on_observations(
                 X=self.transform_inputs(X), Y=Y_fantasized, **kwargs
             )
@@ -362,13 +425,11 @@ class ModelList(Model):
         super().__init__()
         self.models = ModuleList(models)
 
-    def _get_group_subset_indices(
-        self, idcs: Optional[List[int]]
-    ) -> Dict[int, List[int]]:
+    def _get_group_subset_indices(self, idcs: list[int] | None) -> dict[int, list[int]]:
         r"""Convert global subset indices to indices for the individual models.
 
         Args:
-            idcs: A list of inidices to which the `ModelList` model is to be
+            idcs: A list of indices to which the `ModelList` model is to be
                 subset to.
 
         Returns:
@@ -380,9 +441,9 @@ class ModelList(Model):
         output_sizes = [model.num_outputs for model in self.models]
         cum_output_sizes = np.cumsum(output_sizes)
         idcs = [idx % cum_output_sizes[-1] for idx in idcs]
-        group_indices: Dict[int, List[int]] = defaultdict(list)
+        group_indices: dict[int, list[int]] = defaultdict(list)
         for idx in idcs:
-            grp_idx = int(np.argwhere(idx < cum_output_sizes)[0])
+            grp_idx = np.argwhere(idx < cum_output_sizes)[0].item()
             sub_idx = idx - int(np.sum(output_sizes[:grp_idx]))
             group_indices[grp_idx].append(sub_idx)
         return group_indices
@@ -390,10 +451,9 @@ class ModelList(Model):
     def posterior(
         self,
         X: Tensor,
-        output_indices: Optional[List[int]] = None,
-        observation_noise: bool = False,
-        posterior_transform: Optional[Callable[[PosteriorList], Posterior]] = None,
-        **kwargs: Any,
+        output_indices: list[int] | None = None,
+        observation_noise: bool | Tensor = False,
+        posterior_transform: Callable[[PosteriorList], Posterior] | None = None,
     ) -> Posterior:
         r"""Computes the posterior over model outputs at the provided points.
 
@@ -410,7 +470,13 @@ class ModelList(Model):
                 Can be used to speed up computation if only a subset of the
                 model's outputs are required for optimization. If omitted,
                 computes the posterior over all model outputs.
-            observation_noise: If True, add observation noise to the posterior.
+            observation_noise: If True, add the observation noise from the
+                respective likelihoods to the posterior. If a Tensor of shape
+                `(batch_shape) x q x m`, use it directly as the observation
+                noise (with `observation_noise[...,i]` added to the posterior
+                of the `i`-th model). `observation_noise` is assumed
+                to be in the outcome-transformed space, if an outcome transform
+                is used by the model.
             posterior_transform: An optional PosteriorTransform.
 
         Returns:
@@ -418,12 +484,21 @@ class ModelList(Model):
             over `q` points and `m` outputs each.
         """
         group_indices = self._get_group_subset_indices(idcs=output_indices)
-        posteriors = [
-            self.models[i].posterior(
-                X=X, output_indices=idcs, observation_noise=observation_noise
+        posteriors = []
+        for i, idcs in group_indices.items():
+            if isinstance(observation_noise, Tensor):
+                if idcs is None:
+                    start_idx = sum(m.num_outputs for m in self.models[:i])
+                    end_idx = start_idx + self.models[i].num_outputs
+                    idcs = list(range(start_idx, end_idx))
+                obs_noise = observation_noise[..., idcs]
+            else:
+                obs_noise = observation_noise
+            posteriors.append(
+                self.models[i].posterior(
+                    X=X, output_indices=idcs, observation_noise=obs_noise
+                )
             )
-            for i, idcs in group_indices.items()
-        ]
         posterior = PosteriorList(*posteriors)
         if posterior_transform is not None:
             posterior = posterior_transform(posterior)
@@ -457,7 +532,7 @@ class ModelList(Model):
         """
         return sum(model.num_outputs for model in self.models)
 
-    def subset_output(self, idcs: List[int]) -> Model:
+    def subset_output(self, idcs: list[int]) -> Model:
         r"""Subset the model along the output dimension.
 
         Args:
@@ -477,15 +552,17 @@ class ModelList(Model):
         the model `m1` subset to its second output.
         """
         group_indices = self._get_group_subset_indices(idcs=idcs)
-        subset_models = [
-            deepcopy(self.models[grp_idx].subset_output(idcs=sub_idcs))
-            for grp_idx, sub_idcs in group_indices.items()
-        ]
+        subset_models = []
+        for grp_idx, sub_idcs in group_indices.items():
+            subset_model = self.models[grp_idx]
+            if sub_idcs is not None and subset_model.num_outputs != len(sub_idcs):
+                subset_model = subset_model.subset_output(idcs=sub_idcs)
+            subset_models.append(subset_model)
         if len(subset_models) == 1:
             return subset_models[0]
         return self.__class__(*subset_models)
 
-    def transform_inputs(self, X: Tensor) -> List[Tensor]:
+    def transform_inputs(self, X: Tensor) -> list[Tensor]:
         r"""Individually transform the inputs for each model.
 
         Args:
@@ -516,6 +593,86 @@ class ModelList(Model):
                 m.load_state_dict(filtered_dict)
         super().load_state_dict(state_dict=state_dict, strict=strict)
 
+    def fantasize(
+        self,
+        X: Tensor,
+        sampler: MCSampler,
+        observation_noise: Tensor | None = None,
+        evaluation_mask: Tensor | None = None,
+        **kwargs: Any,
+    ) -> Model:
+        r"""Construct a fantasy model.
+
+        Constructs a fantasy model in the following fashion:
+        (1) compute the model posterior at `X` (including observation noise if
+        `observation_noise=True`).
+        (2) sample from this posterior (using `sampler`) to generate "fake"
+        observations.
+        (3) condition the model on the new fake observations.
+
+        Args:
+            X: A `batch_shape x n' x d`-dim Tensor, where `d` is the dimension of
+                the feature space, `n'` is the number of points per batch, and
+                `batch_shape` is the batch shape (must be compatible with the
+                batch shape of the model).
+            sampler: The sampler used for sampling from the posterior at `X`. If
+                evaluation_mask is not None, this must be a `ListSampler`.
+            observation_noise: A `model_batch_shape x 1 x m`-dim tensor or
+                a `model_batch_shape x n' x m`-dim tensor containing the average
+                noise for each batch and output, where `m` is the number of outputs.
+                `noise` must be in the outcome-transformed space if an outcome
+                transform is used. If None, then the noise will be the inferred
+                noise level.
+            evaluation_mask: A `n' x m`-dim tensor of booleans indicating which
+                outputs should be fantasized for a given design. This uses the same
+                evaluation mask for all batches.
+
+        Returns:
+            The constructed fantasy model.
+        """
+        if evaluation_mask is not None:
+            if evaluation_mask.ndim != 2 or evaluation_mask.shape != torch.Size(
+                [X.shape[-2], self.num_outputs]
+            ):
+                raise BotorchTensorDimensionError(
+                    f"Expected evaluation_mask of shape `{X.shape[0]} "
+                    f"x {self.num_outputs}`, but got "
+                    f"{shape_to_str(evaluation_mask.shape)}."
+                )
+            if not isinstance(sampler, ListSampler):
+                raise ValueError("Decoupled fantasization requires a list of samplers.")
+
+        fant_models = []
+        X_i = X
+        if observation_noise is None:
+            observation_noise_i = observation_noise
+        for i in range(self.num_outputs):
+            # get the inputs to fantasize at for output i
+            if evaluation_mask is not None:
+                mask_i = evaluation_mask[:, i]
+                X_i = X[..., mask_i, :]
+                # TODO (T158701749): implement a QMC DecoupledSampler that draws all
+                # samples from a single Sobol sequence or consider requiring that the
+                # sampling is IID to ensure good coverage.
+                sampler_i = sampler.samplers[i]
+                if observation_noise is not None:
+                    observation_noise_i = observation_noise[..., mask_i, i : i + 1]
+            else:
+                sampler_i = (
+                    sampler.samplers[i] if isinstance(sampler, ListSampler) else sampler
+                )
+                if observation_noise is not None:
+                    observation_noise_i = observation_noise[..., i : i + 1]
+
+            fant_model = self.models[i].fantasize(
+                X=X_i,
+                sampler=sampler_i,
+                observation_noise=observation_noise_i,
+                **kwargs,
+            )
+            fant_models.append(fant_model)
+        return self.__class__(*fant_models)
+
 
 class ModelDict(ModuleDict):
     r"""A lightweight container mapping model names to models."""
