@@ -24,13 +24,14 @@ from __future__ import annotations
 
 from abc import ABC, abstractmethod
 from collections import OrderedDict
-from typing import List, Optional, Tuple, Union
 
 import torch
 from botorch.models.transforms.utils import (
+    nanstd,
     norm_to_lognorm_mean,
     norm_to_lognorm_variance,
 )
+from botorch.models.utils.assorted import get_task_value_remapping
 from botorch.posteriors import GPyTorchPosterior, Posterior, TransformedPosterior
 from botorch.utils.transforms import normalize_indices
 from linear_operator.operators import CholLinearOperator, DiagLinearOperator
@@ -39,22 +40,19 @@ from torch.nn import Module, ModuleDict
 
 
 class OutcomeTransform(Module, ABC):
-    r"""
-    Abstract base class for outcome transforms.
-
-    :meta private:
-    """
+    """Abstract base class for outcome transforms."""
 
     @abstractmethod
     def forward(
-        self, Y: Tensor, Yvar: Optional[Tensor] = None
-    ) -> Tuple[Tensor, Optional[Tensor]]:
+        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
+    ) -> tuple[Tensor, Tensor | None]:
         r"""Transform the outcomes in a model's training targets
 
         Args:
             Y: A `batch_shape x n x m`-dim tensor of training targets.
             Yvar: A `batch_shape x n x m`-dim tensor of observation noises
                 associated with the training targets (if applicable).
+            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).
 
         Returns:
             A two-tuple with the transformed outcomes:
@@ -64,7 +62,7 @@ class OutcomeTransform(Module, ABC):
         """
         pass  # pragma: no cover
 
-    def subset_output(self, idcs: List[int]) -> OutcomeTransform:
+    def subset_output(self, idcs: list[int]) -> OutcomeTransform:
         r"""Subset the transform along the output dimension.
 
         This functionality is used to properly treat outcome transformations
@@ -82,14 +80,15 @@ class OutcomeTransform(Module, ABC):
         )
 
     def untransform(
-        self, Y: Tensor, Yvar: Optional[Tensor] = None
-    ) -> Tuple[Tensor, Optional[Tensor]]:
+        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
+    ) -> tuple[Tensor, Tensor | None]:
         r"""Un-transform previously transformed outcomes
 
         Args:
             Y: A `batch_shape x n x m`-dim tensor of transfomred training targets.
             Yvar: A `batch_shape x n x m`-dim tensor of transformed observation
                 noises associated with the training targets (if applicable).
+            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).
 
         Returns:
             A two-tuple with the un-transformed outcomes:
@@ -111,7 +110,9 @@ class OutcomeTransform(Module, ABC):
         """
         return False
 
-    def untransform_posterior(self, posterior: Posterior) -> Posterior:
+    def untransform_posterior(
+        self, posterior: Posterior, X: Tensor | None = None
+    ) -> Posterior:
         r"""Un-transform a posterior.
 
         Posteriors with `_is_linear=True` should return a `GPyTorchPosterior` when
@@ -120,6 +121,7 @@ class OutcomeTransform(Module, ABC):
 
         Args:
             posterior: A posterior in the transformed space.
+            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).
 
         Returns:
             The un-transformed posterior.
@@ -144,14 +146,15 @@ class ChainedOutcomeTransform(OutcomeTransform, ModuleDict):
         super().__init__(OrderedDict(transforms))
 
     def forward(
-        self, Y: Tensor, Yvar: Optional[Tensor] = None
-    ) -> Tuple[Tensor, Optional[Tensor]]:
+        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
+    ) -> tuple[Tensor, Tensor | None]:
         r"""Transform the outcomes in a model's training targets
 
         Args:
             Y: A `batch_shape x n x m`-dim tensor of training targets.
             Yvar: A `batch_shape x n x m`-dim tensor of observation noises
                 associated with the training targets (if applicable).
+            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).
 
         Returns:
             A two-tuple with the transformed outcomes:
@@ -160,10 +163,10 @@ class ChainedOutcomeTransform(OutcomeTransform, ModuleDict):
             - The transformed observation noise (if applicable).
         """
         for tf in self.values():
-            Y, Yvar = tf.forward(Y, Yvar)
+            Y, Yvar = tf.forward(Y=Y, Yvar=Yvar, X=X)
         return Y, Yvar
 
-    def subset_output(self, idcs: List[int]) -> OutcomeTransform:
+    def subset_output(self, idcs: list[int]) -> OutcomeTransform:
         r"""Subset the transform along the output dimension.
 
         Args:
@@ -177,14 +180,15 @@ class ChainedOutcomeTransform(OutcomeTransform, ModuleDict):
         )
 
     def untransform(
-        self, Y: Tensor, Yvar: Optional[Tensor] = None
-    ) -> Tuple[Tensor, Optional[Tensor]]:
+        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
+    ) -> tuple[Tensor, Tensor | None]:
         r"""Un-transform previously transformed outcomes
 
         Args:
             Y: A `batch_shape x n x m`-dim tensor of transfomred training targets.
             Yvar: A `batch_shape x n x m`-dim tensor of transformed observation
                 noises associated with the training targets (if applicable).
+            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).
 
         Returns:
             A two-tuple with the un-transformed outcomes:
@@ -193,7 +197,7 @@ class ChainedOutcomeTransform(OutcomeTransform, ModuleDict):
             - The un-transformed observation noise (if applicable).
         """
         for tf in reversed(self.values()):
-            Y, Yvar = tf.untransform(Y, Yvar)
+            Y, Yvar = tf.untransform(Y=Y, Yvar=Yvar, X=X)
         return Y, Yvar
 
     @property
@@ -202,19 +206,22 @@ class ChainedOutcomeTransform(OutcomeTransform, ModuleDict):
         A `ChainedOutcomeTransform` is linear only if all of the component transforms
         are linear.
         """
-        return all((octf._is_linear for octf in self.values()))
+        return all(octf._is_linear for octf in self.values())
 
-    def untransform_posterior(self, posterior: Posterior) -> Posterior:
+    def untransform_posterior(
+        self, posterior: Posterior, X: Tensor | None = None
+    ) -> Posterior:
         r"""Un-transform a posterior
 
         Args:
             posterior: A posterior in the transformed space.
+            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).
 
         Returns:
             The un-transformed posterior.
         """
         for tf in reversed(self.values()):
-            posterior = tf.untransform_posterior(posterior)
+            posterior = tf.untransform_posterior(posterior, X=X)
         return posterior
 
 
@@ -230,7 +237,7 @@ class Standardize(OutcomeTransform):
     def __init__(
         self,
         m: int,
-        outputs: Optional[List[int]] = None,
+        outputs: list[int] | None = None,
         batch_shape: torch.Size = torch.Size(),  # noqa: B008
         min_stdv: float = 1e-8,
     ) -> None:
@@ -245,17 +252,58 @@ class Standardize(OutcomeTransform):
                 standardization (if lower, only de-mean the data).
         """
         super().__init__()
-        self.register_buffer("means", None)
-        self.register_buffer("stdvs", None)
-        self.register_buffer("_stdvs_sq", None)
+        self.register_buffer("means", torch.zeros(*batch_shape, 1, m))
+        self.register_buffer("stdvs", torch.ones(*batch_shape, 1, m))
+        self.register_buffer("_stdvs_sq", torch.ones(*batch_shape, 1, m))
+        self.register_buffer("_is_trained", torch.tensor(False))
         self._outputs = normalize_indices(outputs, d=m)
         self._m = m
         self._batch_shape = batch_shape
         self._min_stdv = min_stdv
 
+    def _get_per_input_means_stdvs(
+        self, X: Tensor, include_stdvs_sq: bool
+    ) -> tuple[Tensor, Tensor, Tensor | None]:
+        r"""Get per-input means and stdvs.
+
+        Args:
+            X: A `batch_shape x n x d`-dim tensor of input parameters.
+            include_stdvs_sq: Whether to include the stdvs squared.
+                This parameter is not used by this method
+
+        Returns:
+            A three-tuple with the  means and stdvs:
+
+            - The per-input means.
+            - The per-input stdvs.
+            - The per-input stdvs squared.
+        """
+        return self.means, self.stdvs, self._stdvs_sq
+
+    def _validate_training_inputs(self, Y: Tensor, Yvar: Tensor | None = None) -> None:
+        """Validate training inputs.
+
+        Args:
+            Y: A `batch_shape x n x m`-dim tensor of training targets.
+            Yvar: A `batch_shape x n x m`-dim tensor of observation noises.
+        """
+        if Y.shape[:-2] != self._batch_shape:
+            raise RuntimeError(
+                f"Expected Y.shape[:-2] to be {self._batch_shape}, matching "
+                f"the `batch_shape` argument to `{self.__class__.__name__}`, but got "
+                f"Y.shape[:-2]={Y.shape[:-2]}."
+            )
+        elif Y.shape[-2] < 1:
+            raise ValueError(f"Can't standardize with no observations. {Y.shape=}.")
+        elif Y.size(-1) != self._m:
+            raise RuntimeError(
+                f"Wrong output dimension. Y.size(-1) is {Y.size(-1)}; expected "
+                f"{self._m}."
+            )
+
     def forward(
-        self, Y: Tensor, Yvar: Optional[Tensor] = None
-    ) -> Tuple[Tensor, Optional[Tensor]]:
+        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
+    ) -> tuple[Tensor, Tensor | None]:
         r"""Standardize outcomes.
 
         If the module is in train mode, this updates the module state (i.e. the
@@ -266,6 +314,9 @@ class Standardize(OutcomeTransform):
             Y: A `batch_shape x n x m`-dim tensor of training targets.
             Yvar: A `batch_shape x n x m`-dim tensor of observation noises
                 associated with the training targets (if applicable).
+            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).
+                This argument is not used by this transform, but it is used by
+                its subclass, `StratifiedStandardize`.
 
         Returns:
             A two-tuple with the transformed outcomes:
@@ -274,14 +325,13 @@ class Standardize(OutcomeTransform):
             - The transformed observation noise (if applicable).
         """
         if self.training:
-            if Y.shape[:-2] != self._batch_shape:
-                raise RuntimeError("wrong batch shape")
-            if Y.size(-1) != self._m:
-                raise RuntimeError(
-                    f"Wrong output dimension. Y.size(-1) is {Y.size(-1)}; expected "
-                    f"{self._m}."
+            self._validate_training_inputs(Y=Y, Yvar=Yvar)
+            if Y.shape[-2] == 1:
+                stdvs = torch.ones(
+                    (*Y.shape[:-2], 1, Y.shape[-1]), dtype=Y.dtype, device=Y.device
                 )
-            stdvs = Y.std(dim=-2, keepdim=True)
+            else:
+                stdvs = Y.std(dim=-2, keepdim=True)
             stdvs = stdvs.where(stdvs >= self._min_stdv, torch.full_like(stdvs, 1.0))
             means = Y.mean(dim=-2, keepdim=True)
             if self._outputs is not None:
@@ -291,12 +341,16 @@ class Standardize(OutcomeTransform):
             self.means = means
             self.stdvs = stdvs
             self._stdvs_sq = stdvs.pow(2)
-
-        Y_tf = (Y - self.means) / self.stdvs
-        Yvar_tf = Yvar / self._stdvs_sq if Yvar is not None else None
+            self._is_trained = torch.tensor(True)
+        include_stdvs_sq = Yvar is not None
+        means, stdvs, stdvs_sq = self._get_per_input_means_stdvs(
+            X=X, include_stdvs_sq=include_stdvs_sq
+        )
+        Y_tf = (Y - means) / stdvs
+        Yvar_tf = Yvar / stdvs_sq if include_stdvs_sq else None
         return Y_tf, Yvar_tf
 
-    def subset_output(self, idcs: List[int]) -> OutcomeTransform:
+    def subset_output(self, idcs: list[int]) -> OutcomeTransform:
         r"""Subset the transform along the output dimension.
 
         Args:
@@ -321,23 +375,26 @@ class Standardize(OutcomeTransform):
             batch_shape=self._batch_shape,
             min_stdv=self._min_stdv,
         )
-        if self.means is not None:
-            new_tf.means = self.means[..., nlzd_idcs]
-            new_tf.stdvs = self.stdvs[..., nlzd_idcs]
-            new_tf._stdvs_sq = self._stdvs_sq[..., nlzd_idcs]
+        new_tf.means = self.means[..., nlzd_idcs]
+        new_tf.stdvs = self.stdvs[..., nlzd_idcs]
+        new_tf._stdvs_sq = self._stdvs_sq[..., nlzd_idcs]
+        new_tf._is_trained = self._is_trained
         if not self.training:
             new_tf.eval()
         return new_tf
 
     def untransform(
-        self, Y: Tensor, Yvar: Optional[Tensor] = None
-    ) -> Tuple[Tensor, Optional[Tensor]]:
+        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
+    ) -> tuple[Tensor, Tensor | None]:
         r"""Un-standardize outcomes.
 
         Args:
             Y: A `batch_shape x n x m`-dim tensor of standardized targets.
             Yvar: A `batch_shape x n x m`-dim tensor of standardized observation
                 noises associated with the targets (if applicable).
+            X: A `batch_shape x n x d`-dim tensor of inputs (if applicable).
+                This argument is not used by this transform, but it is used by
+                its subclass, `StratifiedStandardize`.
 
         Returns:
             A two-tuple with the un-standardized outcomes:
@@ -345,15 +402,18 @@ class Standardize(OutcomeTransform):
             - The un-standardized outcome observations.
             - The un-standardized observation noise (if applicable).
         """
-        if self.means is None:
+        if not self._is_trained:
             raise RuntimeError(
                 "`Standardize` transforms must be called on outcome data "
                 "(e.g. `transform(Y)`) before calling `untransform`, since "
                 "means and standard deviations need to be computed."
             )
-
-        Y_utf = self.means + self.stdvs * Y
-        Yvar_utf = self._stdvs_sq * Yvar if Yvar is not None else None
+        include_stdvs_sq = Yvar is not None
+        means, stdvs, stdvs_sq = self._get_per_input_means_stdvs(
+            X=X, include_stdvs_sq=include_stdvs_sq
+        )
+        Y_utf = means + stdvs * Y
+        Yvar_utf = stdvs_sq * Yvar if include_stdvs_sq else None
         return Y_utf, Yvar_utf
 
     @property
@@ -361,12 +421,15 @@ class Standardize(OutcomeTransform):
         return True
 
     def untransform_posterior(
-        self, posterior: Posterior
-    ) -> Union[GPyTorchPosterior, TransformedPosterior]:
+        self, posterior: Posterior, X: Tensor | None = None
+    ) -> GPyTorchPosterior | TransformedPosterior:
         r"""Un-standardize the posterior.
 
         Args:
             posterior: A posterior in the standardized space.
+            X: A `batch_shape x n x d`-dim tensor of inputs (if applicable).
+                This argument is not used by this transform, but it is used by
+                its subclass, `StratifiedStandardize`.
 
         Returns:
             The un-standardized posterior. If the input posterior is a
@@ -378,7 +441,7 @@ class Standardize(OutcomeTransform):
                 "Standardize does not yet support output selection for "
                 "untransform_posterior"
             )
-        if self.means is None:
+        if not self._is_trained:
             raise RuntimeError(
                 "`Standardize` transforms must be called on outcome data "
                 "(e.g. `transform(Y)`) before calling `untransform_posterior`, since "
@@ -405,8 +468,9 @@ class Standardize(OutcomeTransform):
             )
         # GPyTorchPosterior (TODO: Should we Lazy-evaluate the mean here as well?)
         mvn = posterior.distribution
-        offset = self.means
-        scale_fac = self.stdvs
+        offset, scale_fac, _ = self._get_per_input_means_stdvs(
+            X=X, include_stdvs_sq=False
+        )
         if not posterior._is_mt:
             mean_tf = offset.squeeze(-1) + scale_fac.squeeze(-1) * mvn.mean
             scale_fac = scale_fac.squeeze(-1).expand_as(mean_tf)
@@ -421,7 +485,6 @@ class Standardize(OutcomeTransform):
 
         if (
             not mvn.islazy
-            # TODO: Figure out attribute namming weirdness here
             or mvn._MultivariateNormal__unbroadcasted_scale_tril is not None
         ):
             # if already computed, we can save a lot of time using scale_tril
@@ -437,6 +500,197 @@ class Standardize(OutcomeTransform):
         return GPyTorchPosterior(mvn_tf)
 
 
+class StratifiedStandardize(Standardize):
+    r"""Standardize outcomes (zero mean, unit variance) along stratification dimension.
+
+    This module is stateful: If in train mode, calling forward updates the
+    module state (i.e. the mean/std normalizing constants). If in eval mode,
+    calling forward simply applies the standardization using the current module
+    state.
+    """
+
+    def __init__(
+        self,
+        task_values: Tensor,
+        stratification_idx: int,
+        batch_shape: torch.Size = torch.Size(),  # noqa: B008
+        min_stdv: float = 1e-8,
+        # dtype: torch.dtype = torch.double,
+    ) -> None:
+        r"""Standardize outcomes (zero mean, unit variance) along stratification dim.
+
+        Note: This currenlty only supports single output models
+        (including multi-task models that have a single output).
+
+        Args:
+            task_values: `t`-dim tensor of task values.
+            stratification_idx: The index of the stratification dimension.
+            batch_shape: The batch_shape of the training targets.
+            min_stddv: The minimum standard deviation for which to perform
+                standardization (if lower, only de-mean the data).
+        """
+        OutcomeTransform.__init__(self)
+        self._stratification_idx = stratification_idx
+        task_values = task_values.unique(sorted=True)
+        self.strata_mapping = get_task_value_remapping(task_values, dtype=torch.double)
+        if self.strata_mapping is None:
+            self.strata_mapping = task_values
+        n_strata = self.strata_mapping.shape[0]
+        self._min_stdv = min_stdv
+        self.register_buffer("means", torch.zeros(*batch_shape, n_strata, 1))
+        self.register_buffer("stdvs", torch.ones(*batch_shape, n_strata, 1))
+        self.register_buffer("_stdvs_sq", torch.ones(*batch_shape, n_strata, 1))
+        self.register_buffer("_is_trained", torch.tensor(False))
+        self._batch_shape = batch_shape
+        self._m = 1  # TODO: support multiple outputs
+        self._outputs = None
+
+    def forward(
+        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
+    ) -> tuple[Tensor, Tensor | None]:
+        r"""Standardize outcomes.
+
+        If the module is in train mode, this updates the module state (i.e. the
+        mean/std normalizing constants). If the module is in eval mode, simply
+        applies the normalization using the module state.
+
+        Args:
+            Y: A `batch_shape x n x m`-dim tensor of training targets.
+            Yvar: A `batch_shape x n x m`-dim tensor of observation noises
+                associated with the training targets (if applicable).
+            X: A `batch_shape x n x d`-dim tensor of input parameters.
+
+        Returns:
+            A two-tuple with the transformed outcomes:
+
+            - The transformed outcome observations.
+            - The transformed observation noise (if applicable).
+        """
+        if X is None:
+            raise ValueError("X is required for StratifiedStandardize.")
+        if self.training:
+            self._validate_training_inputs(Y=Y, Yvar=Yvar)
+            self.means = self.means.to(dtype=X.dtype, device=X.device)
+            self.stdvs = self.stdvs.to(dtype=X.dtype, device=X.device)
+            self._stdvs_sq = self._stdvs_sq.to(dtype=X.dtype, device=X.device)
+            strata = X[..., self._stratification_idx].long()
+            unique_strata = strata.unique()
+            for s in unique_strata:
+                mapped_strata = self.strata_mapping[s].long()
+                mask = strata != s
+                Y_strata = Y.clone()
+                Y_strata[..., mask, :] = float("nan")
+                stdvs = (
+                    torch.ones_like(Y_strata)
+                    if Y.shape[-2] == 1
+                    else nanstd(X=Y_strata, dim=-2)
+                )
+                stdvs = stdvs.where(
+                    stdvs >= self._min_stdv, torch.full_like(stdvs, 1.0)
+                )
+                means = Y_strata.nanmean(dim=-2)
+                self.means[..., mapped_strata, :] = means
+                self.stdvs[..., mapped_strata, :] = stdvs
+                self._stdvs_sq[..., mapped_strata, :] = stdvs.pow(2)
+            self._is_trained = torch.tensor(True)
+        training = self.training
+        self.training = False
+        tf_Y, tf_Yvar = super().forward(Y=Y, Yvar=Yvar, X=X)
+        self.training = training
+        return tf_Y, tf_Yvar
+
+    def _get_per_input_means_stdvs(
+        self, X: Tensor, include_stdvs_sq: bool
+    ) -> tuple[Tensor, Tensor, Tensor | None]:
+        r"""Get per-input means and stdvs.
+
+        Args:
+            X: A `batch_shape x n x d`-dim tensor of input parameters.
+            include_stdvs_sq: Whether to include the stdvs squared.
+
+        Returns:
+            A three-tuple with the per-input means and stdvs:
+
+            - The per-input means.
+            - The per-input stdvs.
+            - The per-input stdvs squared.
+        """
+        strata = X[..., self._stratification_idx].long()
+        mapped_strata = self.strata_mapping[strata].unsqueeze(-1).long()
+        # get means and stdvs for each strata
+        n_extra_batch_dims = mapped_strata.ndim - 2 - len(self._batch_shape)
+        expand_shape = mapped_strata.shape[:n_extra_batch_dims] + self.means.shape
+        means = torch.gather(
+            input=self.means.expand(expand_shape),
+            dim=-2,
+            index=mapped_strata,
+        )
+        stdvs = torch.gather(
+            input=self.stdvs.expand(expand_shape),
+            dim=-2,
+            index=mapped_strata,
+        )
+        if include_stdvs_sq:
+            stdvs_sq = torch.gather(
+                input=self._stdvs_sq.expand(expand_shape),
+                dim=-2,
+                index=mapped_strata,
+            )
+        else:
+            stdvs_sq = None
+        return means, stdvs, stdvs_sq
+
+    def subset_output(self, idcs: list[int]) -> OutcomeTransform:
+        r"""Subset the transform along the output dimension.
+
+        Args:
+            idcs: The output indices to subset the transform to.
+
+        Returns:
+            The current outcome transform, subset to the specified output indices.
+        """
+        raise NotImplementedError
+
+    def untransform(
+        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
+    ) -> tuple[Tensor, Tensor | None]:
+        r"""Un-standardize outcomes.
+
+        Args:
+            Y: A `batch_shape x n x m`-dim tensor of standardized targets.
+            Yvar: A `batch_shape x n x m`-dim tensor of standardized observation
+                noises associated with the targets (if applicable).
+            X: A `batch_shape x n x d`-dim tensor of input parameters.
+
+        Returns:
+            A two-tuple with the un-standardized outcomes:
+
+            - The un-standardized outcome observations.
+            - The un-standardized observation noise (if applicable).
+        """
+        if X is None:
+            raise ValueError("X is required for StratifiedStandardize.")
+        return super().untransform(Y=Y, Yvar=Yvar, X=X)
+
+    def untransform_posterior(
+        self, posterior: Posterior, X: Tensor | None = None
+    ) -> GPyTorchPosterior | TransformedPosterior:
+        r"""Un-standardize the posterior.
+
+        Args:
+            posterior: A posterior in the standardized space.
+            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).
+
+        Returns:
+            The un-standardized posterior. If the input posterior is a
+            `GPyTorchPosterior`, return a `GPyTorchPosterior`. Otherwise, return a
+            `TransformedPosterior`.
+        """
+        if X is None:
+            raise ValueError("X is required for StratifiedStandardize.")
+        return super().untransform_posterior(posterior=posterior, X=X)
+
+
 class Log(OutcomeTransform):
     r"""Log-transform outcomes.
 
@@ -445,7 +699,7 @@ class Log(OutcomeTransform):
     log-transformed outcomes and un-transform the model posterior of that GP.
     """
 
-    def __init__(self, outputs: Optional[List[int]] = None) -> None:
+    def __init__(self, outputs: list[int] | None = None) -> None:
         r"""Log-transform outcomes.
 
         Args:
@@ -455,7 +709,7 @@ class Log(OutcomeTransform):
         super().__init__()
         self._outputs = outputs
 
-    def subset_output(self, idcs: List[int]) -> OutcomeTransform:
+    def subset_output(self, idcs: list[int]) -> OutcomeTransform:
         r"""Subset the transform along the output dimension.
 
         Args:
@@ -478,14 +732,16 @@ class Log(OutcomeTransform):
         return new_tf
 
     def forward(
-        self, Y: Tensor, Yvar: Optional[Tensor] = None
-    ) -> Tuple[Tensor, Optional[Tensor]]:
+        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
+    ) -> tuple[Tensor, Tensor | None]:
         r"""Log-transform outcomes.
 
         Args:
             Y: A `batch_shape x n x m`-dim tensor of training targets.
             Yvar: A `batch_shape x n x m`-dim tensor of observation noises
                 associated with the training targets (if applicable).
+            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).
+                This argument is not used by this transform.
 
         Returns:
             A two-tuple with the transformed outcomes:
@@ -511,8 +767,8 @@ class Log(OutcomeTransform):
         return Y_tf, Yvar
 
     def untransform(
-        self, Y: Tensor, Yvar: Optional[Tensor] = None
-    ) -> Tuple[Tensor, Optional[Tensor]]:
+        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
+    ) -> tuple[Tensor, Tensor | None]:
         r"""Un-transform log-transformed outcomes
 
         Args:
@@ -520,6 +776,8 @@ class Log(OutcomeTransform):
             Yvar: A `batch_shape x n x m`-dim tensor of log- transformed
                 observation noises associated with the training targets
                 (if applicable).
+            X: A `batch_shape x n x d`-dim tensor of inputs (if applicable).
+                This argument is not used by this transform.
 
         Returns:
             A two-tuple with the un-transformed outcomes:
@@ -544,11 +802,15 @@ class Log(OutcomeTransform):
             )
         return Y_utf, Yvar
 
-    def untransform_posterior(self, posterior: Posterior) -> TransformedPosterior:
+    def untransform_posterior(
+        self, posterior: Posterior, X: Tensor | None = None
+    ) -> TransformedPosterior:
         r"""Un-transform the log-transformed posterior.
 
         Args:
             posterior: A posterior in the log-transformed space.
+            X: A `batch_shape x n x d`-dim tensor of inputs (if applicable).
+                This argument is not used by this transform.
 
         Returns:
             The un-transformed posterior.
@@ -573,7 +835,7 @@ class Power(OutcomeTransform):
     power-transformed outcomes and un-transform the model posterior of that GP.
     """
 
-    def __init__(self, power: float, outputs: Optional[List[int]] = None) -> None:
+    def __init__(self, power: float, outputs: list[int] | None = None) -> None:
         r"""Power-transform outcomes.
 
         Args:
@@ -584,7 +846,7 @@ class Power(OutcomeTransform):
         self._outputs = outputs
         self.power = power
 
-    def subset_output(self, idcs: List[int]) -> OutcomeTransform:
+    def subset_output(self, idcs: list[int]) -> OutcomeTransform:
         r"""Subset the transform along the output dimension.
 
         Args:
@@ -607,14 +869,16 @@ class Power(OutcomeTransform):
         return new_tf
 
     def forward(
-        self, Y: Tensor, Yvar: Optional[Tensor] = None
-    ) -> Tuple[Tensor, Optional[Tensor]]:
+        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
+    ) -> tuple[Tensor, Tensor | None]:
         r"""Power-transform outcomes.
 
         Args:
             Y: A `batch_shape x n x m`-dim tensor of training targets.
             Yvar: A `batch_shape x n x m`-dim tensor of observation noises
                 associated with the training targets (if applicable).
+            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).
+                This argument is not used by this transform.
 
         Returns:
             A two-tuple with the transformed outcomes:
@@ -640,8 +904,8 @@ class Power(OutcomeTransform):
         return Y_tf, Yvar
 
     def untransform(
-        self, Y: Tensor, Yvar: Optional[Tensor] = None
-    ) -> Tuple[Tensor, Optional[Tensor]]:
+        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
+    ) -> tuple[Tensor, Tensor | None]:
         r"""Un-transform power-transformed outcomes
 
         Args:
@@ -649,6 +913,8 @@ class Power(OutcomeTransform):
             Yvar: A `batch_shape x n x m`-dim tensor of power-transformed
                 observation noises associated with the training targets
                 (if applicable).
+            X: A `batch_shape x n x d`-dim tensor of inputs (if applicable).
+                This argument is not used by this transform.
 
         Returns:
             A two-tuple with the un-transformed outcomes:
@@ -673,11 +939,15 @@ class Power(OutcomeTransform):
             )
         return Y_utf, Yvar
 
-    def untransform_posterior(self, posterior: Posterior) -> TransformedPosterior:
+    def untransform_posterior(
+        self, posterior: Posterior, X: Tensor | None = None
+    ) -> TransformedPosterior:
         r"""Un-transform the power-transformed posterior.
 
         Args:
             posterior: A posterior in the power-transformed space.
+            X: A `batch_shape x n x d`-dim tensor of inputs (if applicable).
+                This argument is not used by this transform.
 
         Returns:
             The un-transformed posterior.
@@ -699,7 +969,7 @@ class Bilog(OutcomeTransform):
     constraints as it magnifies values near zero and flattens extreme values.
     """
 
-    def __init__(self, outputs: Optional[List[int]] = None) -> None:
+    def __init__(self, outputs: list[int] | None = None) -> None:
         r"""Bilog-transform outcomes.
 
         Args:
@@ -709,7 +979,7 @@ class Bilog(OutcomeTransform):
         super().__init__()
         self._outputs = outputs
 
-    def subset_output(self, idcs: List[int]) -> OutcomeTransform:
+    def subset_output(self, idcs: list[int]) -> OutcomeTransform:
         r"""Subset the transform along the output dimension.
 
         Args:
@@ -732,14 +1002,16 @@ class Bilog(OutcomeTransform):
         return new_tf
 
     def forward(
-        self, Y: Tensor, Yvar: Optional[Tensor] = None
-    ) -> Tuple[Tensor, Optional[Tensor]]:
+        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
+    ) -> tuple[Tensor, Tensor | None]:
         r"""Bilog-transform outcomes.
 
         Args:
             Y: A `batch_shape x n x m`-dim tensor of training targets.
             Yvar: A `batch_shape x n x m`-dim tensor of observation noises
                 associated with the training targets (if applicable).
+            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).
+                This argument is not used by this transform.
 
         Returns:
             A two-tuple with the transformed outcomes:
@@ -764,8 +1036,8 @@ class Bilog(OutcomeTransform):
         return Y_tf, Yvar
 
     def untransform(
-        self, Y: Tensor, Yvar: Optional[Tensor] = None
-    ) -> Tuple[Tensor, Optional[Tensor]]:
+        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
+    ) -> tuple[Tensor, Tensor | None]:
         r"""Un-transform bilog-transformed outcomes
 
         Args:
@@ -773,6 +1045,8 @@ class Bilog(OutcomeTransform):
             Yvar: A `batch_shape x n x m`-dim tensor of bilog-transformed
                 observation noises associated with the training targets
                 (if applicable).
+            X: A `batch_shape x n x d`-dim tensor of inputs (if applicable).
+                This argument is not used by this transform.
 
         Returns:
             A two-tuple with the un-transformed outcomes:
@@ -780,7 +1054,7 @@ class Bilog(OutcomeTransform):
             - The un-transformed outcome observations.
             - The un-transformed observation noise (if applicable).
         """
-        Y_utf = Y.sign() * (Y.abs().exp() - 1.0)
+        Y_utf = Y.sign() * Y.abs().expm1()
         outputs = normalize_indices(self._outputs, d=Y.size(-1))
         if outputs is not None:
             Y_utf = torch.stack(
@@ -797,11 +1071,15 @@ class Bilog(OutcomeTransform):
             )
         return Y_utf, Yvar
 
-    def untransform_posterior(self, posterior: Posterior) -> TransformedPosterior:
+    def untransform_posterior(
+        self, posterior: Posterior, X: Tensor | None = None
+    ) -> TransformedPosterior:
         r"""Un-transform the bilog-transformed posterior.
 
         Args:
             posterior: A posterior in the bilog-transformed space.
+            X: A `batch_shape x n x d`-dim tensor of inputs (if applicable).
+                This argument is not used by this transform.
 
         Returns:
             The un-transformed posterior.
@@ -812,5 +1090,5 @@ class Bilog(OutcomeTransform):
             )
         return TransformedPosterior(
             posterior=posterior,
-            sample_transform=lambda x: x.sign() * (x.abs().exp() - 1.0),
+            sample_transform=lambda x: x.sign() * x.abs().expm1(),
         )
