@@ -10,27 +10,29 @@ Utility functions for constrained optimization.
 
 from __future__ import annotations
 
+from collections.abc import Callable
 from functools import partial
-from typing import Callable, Dict, List, Optional, Tuple, Union
+from typing import Union
 
 import numpy as np
+import numpy.typing as npt
 import torch
 from botorch.exceptions.errors import CandidateGenerationError, UnsupportedError
 from scipy.optimize import Bounds
 from torch import Tensor
 
 
-ScipyConstraintDict = Dict[
+ScipyConstraintDict = dict[
     str, Union[str, Callable[[np.ndarray], float], Callable[[np.ndarray], np.ndarray]]
 ]
-NLC_TOL = -1e-6
+CONST_TOL = 1e-6
 
 
 def make_scipy_bounds(
     X: Tensor,
-    lower_bounds: Optional[Union[float, Tensor]] = None,
-    upper_bounds: Optional[Union[float, Tensor]] = None,
-) -> Optional[Bounds]:
+    lower_bounds: float | Tensor | None = None,
+    upper_bounds: float | Tensor | None = None,
+) -> Bounds | None:
     r"""Creates a scipy Bounds object for optimziation
 
     Args:
@@ -51,7 +53,7 @@ def make_scipy_bounds(
     if lower_bounds is None and upper_bounds is None:
         return None
 
-    def _expand(bounds: Union[float, Tensor], X: Tensor, lower: bool) -> Tensor:
+    def _expand(bounds: float | Tensor, X: Tensor, lower: bool) -> Tensor:
         if bounds is None:
             ebounds = torch.full_like(X, float("-inf" if lower else "inf"))
         else:
@@ -67,9 +69,9 @@ def make_scipy_bounds(
 
 def make_scipy_linear_constraints(
     shapeX: torch.Size,
-    inequality_constraints: Optional[List[Tuple[Tensor, Tensor, float]]] = None,
-    equality_constraints: Optional[List[Tuple[Tensor, Tensor, float]]] = None,
-) -> List[ScipyConstraintDict]:
+    inequality_constraints: list[tuple[Tensor, Tensor, float]] | None = None,
+    equality_constraints: list[tuple[Tensor, Tensor, float]] | None = None,
+) -> list[ScipyConstraintDict]:
     r"""Generate scipy constraints from torch representation.
 
     Args:
@@ -89,7 +91,7 @@ def make_scipy_linear_constraints(
     Returns:
         A list of dictionaries containing callables for constraint function
         values and Jacobians and a string indicating the associated constraint
-        type ("eq", "ineq"), as expected by `scipy.minimize`.
+        type ("eq", "ineq"), as expected by `scipy.optimize.minimize`.
 
     This function assumes that constraints are the same for each input batch,
     and broadcasts the constraints accordingly to the input batch shape. This
@@ -129,7 +131,7 @@ def make_scipy_linear_constraints(
 
 
 def eval_lin_constraint(
-    x: np.ndarray, flat_idxr: List[int], coeffs: np.ndarray, rhs: float
+    x: npt.NDArray, flat_idxr: list[int], coeffs: npt.NDArray, rhs: float
 ) -> np.float64:
     r"""Evaluate a single linear constraint.
 
@@ -146,8 +148,8 @@ def eval_lin_constraint(
 
 
 def lin_constraint_jac(
-    x: np.ndarray, flat_idxr: List[int], coeffs: np.ndarray, n: int
-) -> np.ndarray:
+    x: npt.NDArray, flat_idxr: list[int], coeffs: npt.NDArray, n: int
+) -> npt.NDArray:
     r"""Return the Jacobian associated with a linear constraint.
 
     Args:
@@ -165,7 +167,7 @@ def lin_constraint_jac(
     return jac
 
 
-def _arrayify(X: Tensor) -> np.ndarray:
+def _arrayify(X: Tensor) -> npt.NDArray:
     r"""Convert a torch.Tensor (any dtype or device) to a numpy (double) array.
 
     Args:
@@ -219,8 +221,8 @@ def _make_linear_constraints(
     rhs: float,
     shapeX: torch.Size,
     eq: bool = False,
-) -> List[ScipyConstraintDict]:
-    r"""Create linear constraints to be used by `scipy.minimize`.
+) -> list[ScipyConstraintDict]:
+    r"""Create linear constraints to be used by `scipy.optimize.minimize`.
 
     Encodes constraints of the form
     `\sum_i (coefficients[i] * X[..., indices[i]]) ? rhs`
@@ -282,7 +284,7 @@ def _make_linear_constraints(
     b, q, d = shapeX
     _validate_linear_constraints_indices_input(indices, q, d)
     n = shapeX.numel()
-    constraints: List[ScipyConstraintDict] = []
+    constraints: list[ScipyConstraintDict] = []
     coeffs = _arrayify(coefficients)
     ctype = "eq" if eq else "ineq"
 
@@ -312,13 +314,109 @@ def _make_linear_constraints(
     return constraints
 
 
+def _make_nonlinear_constraints(
+    f_np_wrapper: Callable, nlc: Callable, is_intrapoint: bool, shapeX: torch.Size
+) -> list[ScipyConstraintDict]:
+    """Create nonlinear constraints to be used by `scipy.optimize.minimize`.
+
+    Args:
+        f_np_wrapper: A wrapper function that given a constraint evaluates
+            the value and gradient (using autograd) of a numpy input and returns both
+            the objective and the gradient.
+        nlc: Callable representing a constraint of the form `callable(x) >= 0`. In case
+            of an intra-point constraint, `callable()`takes in an one-dimensional tensor
+            of shape `d` and returns a scalar. In case of an inter-point constraint,
+            `callable()` takes a two dimensional tensor of shape `q x d` and again
+            returns a scalar.
+        is_intrapoint: A Boolean indicating if a constraint is an intra-point or
+            inter-point constraint (see the docstring of the `inequality_constraints`
+            argument to `optimize_acqf()`).
+        shapeX: Shape of the three-dimensional batch X, that should be optimized.
+
+    Returns:
+        A list of constraint dictionaries with the following keys
+
+        - "type": Indicates the type of the constraint, here always "ineq".
+        - "fun": A callable evaluating the constraint value on `x`, a flattened
+            version of the input tensor `X`, returning a scalar.
+        - "jac": A callable evaluating the constraint's Jacobian on `x`, a flattened
+            version of the input tensor `X`, returning a numpy array.
+    """
+    shapeX = _validate_linear_constraints_shape_input(shapeX)
+    b, q, _ = shapeX
+    constraints = []
+
+    def get_intrapoint_constraint(b: int, q: int, nlc: Callable) -> Callable:
+        return lambda x: nlc(x[b, q])
+
+    def get_interpoint_constraint(b: int, nlc: Callable) -> Callable:
+        return lambda x: nlc(x[b])
+
+    if is_intrapoint:
+        for i in range(b):
+            for j in range(q):
+                f_obj, f_grad = _make_f_and_grad_nonlinear_inequality_constraints(
+                    f_np_wrapper=f_np_wrapper,
+                    nlc=get_intrapoint_constraint(b=i, q=j, nlc=nlc),
+                )
+                constraints.append({"type": "ineq", "fun": f_obj, "jac": f_grad})
+    else:
+        for i in range(b):
+            f_obj, f_grad = _make_f_and_grad_nonlinear_inequality_constraints(
+                f_np_wrapper=f_np_wrapper,
+                nlc=get_interpoint_constraint(b=i, nlc=nlc),
+            )
+            constraints.append({"type": "ineq", "fun": f_obj, "jac": f_grad})
+
+    return constraints
+
+
+def _generate_unfixed_nonlin_constraints(
+    constraints: list[tuple[Callable[[Tensor], Tensor], bool]] | None,
+    fixed_features: dict[int, float],
+    dimension: int,
+) -> list[Callable[[Tensor], Tensor]] | None:
+    """Given a dictionary of fixed features, returns a list of callables for
+    nonlinear inequality constraints expecting only a tensor with the non-fixed
+    features as input.
+    """
+    if not constraints:
+        return constraints
+
+    selector = []
+    idx_X, idx_f = 0, dimension - len(fixed_features)
+    for i in range(dimension):
+        if i in fixed_features.keys():
+            selector.append(idx_f)
+            idx_f += 1
+        else:
+            selector.append(idx_X)
+            idx_X += 1
+
+    values = torch.tensor(list(fixed_features.values()), dtype=torch.double)
+
+    def _wrap_nonlin_constraint(
+        constraint: Callable[[Tensor], Tensor],
+    ) -> Callable[[Tensor], Tensor]:
+        def new_nonlin_constraint(X: Tensor) -> Tensor:
+            ivalues = values.to(X).expand(*X.shape[:-1], len(fixed_features))
+            X_perm = torch.cat([X, ivalues], dim=-1)
+            return constraint(X_perm[..., selector])
+
+        return new_nonlin_constraint
+
+    return [
+        (_wrap_nonlin_constraint(constraint=nlc), is_intrapoint)
+        for nlc, is_intrapoint in constraints
+    ]
+
+
 def _generate_unfixed_lin_constraints(
-    constraints: Optional[List[Tuple[Tensor, Tensor, float]]],
-    fixed_features: Dict[int, float],
+    constraints: list[tuple[Tensor, Tensor, float]] | None,
+    fixed_features: dict[int, float],
     dimension: int,
     eq: bool,
-) -> Optional[List[Tuple[Tensor, Tensor, float]]]:
-
+) -> list[tuple[Tensor, Tensor, float]] | None:
     # If constraints is None or an empty list, then return itself
     if not constraints:
         return constraints
@@ -377,11 +475,11 @@ def _generate_unfixed_lin_constraints(
 
 def _make_f_and_grad_nonlinear_inequality_constraints(
     f_np_wrapper: Callable, nlc: Callable
-) -> Tuple[Callable[[Tensor], Tensor], Callable[[Tensor], Tensor]]:
+) -> tuple[Callable[[Tensor], Tensor], Callable[[Tensor], Tensor]]:
     """
     Create callables for objective + grad for the nonlinear inequality constraints.
     The Scipy interface requires specifying separate callables and we use caching to
-    avoid evaluating the same input twice. This caching onlh works if
+    avoid evaluating the same input twice. This caching only works if
     the returned functions are evaluated on the same input in immediate
     sequence (i.e., calling `f_obj(X_1)`, `f_grad(X_1)` will result in a
     single forward pass, while `f_obj(X_1)`, `f_grad(X_2)`, `f_obj(X_1)`
@@ -411,50 +509,185 @@ def _make_f_and_grad_nonlinear_inequality_constraints(
     return f_obj, f_grad
 
 
+def nonlinear_constraint_is_feasible(
+    nonlinear_inequality_constraint: Callable,
+    is_intrapoint: bool,
+    x: Tensor,
+    tolerance: float = CONST_TOL,
+) -> Tensor:
+    """Checks if a nonlinear inequality constraint is fulfilled (within tolerance).
+
+    Args:
+        nonlinear_inequality_constraint: Callable to evaluate the
+            constraint.
+        intra: If True, the constraint is an intra-point constraint that
+            is applied pointwise and is broadcasted over the q-batch. Else, the
+            constraint has to evaluated over the whole q-batch and is a an
+            inter-point constraint.
+        x: Tensor of shape (batch x q x d).
+        tolerance: Rather than using the exact `const(x) >= 0` constraint, this helper
+            checks feasibility of `const(x) >= -tolerance`. This avoids marking the
+            candidates as infeasible due to tiny violations.
+
+    Returns:
+        A boolean tensor of shape (batch) indicating if the constraint is
+        satified by the corresponding batch of `x`.
+    """
+
+    def check_x(x: Tensor) -> bool:
+        return _arrayify(nonlinear_inequality_constraint(x)).item() >= -tolerance
+
+    x_flat = x.view(-1, *x.shape[-2:])
+    is_feasible = torch.ones(x_flat.shape[0], dtype=torch.bool, device=x.device)
+    for i, x_ in enumerate(x_flat):
+        if is_intrapoint:
+            is_feasible[i] &= all(check_x(x__) for x__ in x_)
+        else:
+            is_feasible[i] &= check_x(x_)
+    return is_feasible.view(x.shape[:-2])
+
+
 def make_scipy_nonlinear_inequality_constraints(
-    nonlinear_inequality_constraints: List[Callable],
+    nonlinear_inequality_constraints: list[tuple[Callable, bool]],
     f_np_wrapper: Callable,
     x0: Tensor,
-) -> List[Dict]:
+    shapeX: torch.Size,
+) -> list[dict]:
     r"""Generate Scipy nonlinear inequality constraints from callables.
 
     Args:
-        nonlinear_inequality_constraints: List of callables for the nonlinear
-            inequality constraints. Each callable represents a constraint of the
-            form >= 0 and takes a torch tensor of size (p x q x dim) and returns a
-            torch tensor of size (p x q).
+        nonlinear_inequality_constraints: A list of tuples representing the nonlinear
+            inequality constraints. The first element in the tuple is a callable
+            representing a constraint of the form `callable(x) >= 0`. In case of an
+            intra-point constraint, `callable()`takes in an one-dimensional tensor of
+            shape `d` and returns a scalar. In case of an inter-point constraint,
+            `callable()` takes a two dimensional tensor of shape `q x d` and again
+            returns a scalar. The second element is a boolean, indicating if it is an
+            intra-point or inter-point constraint (`True` for intra-point. `False` for
+            inter-point). For more information on intra-point vs inter-point
+            constraints, see the docstring of the `inequality_constraints` argument to
+            `optimize_acqf()`. The constraints will later be passed to the scipy
+            solver.
         f_np_wrapper: A wrapper function that given a constraint evaluates the value
              and gradient (using autograd) of a numpy input and returns both the
              objective and the gradient.
         x0: The starting point for SLSQP. We return this starting point in (rare)
             cases where SLSQP fails and thus require it to be feasible.
+        shapeX: Shape of the three-dimensional batch X, that should be optimized.
 
     Returns:
         A list of dictionaries containing callables for constraint function
         values and Jacobians and a string indicating the associated constraint
-        type ("eq", "ineq"), as expected by `scipy.minimize`.
+        type ("eq", "ineq"), as expected by `scipy.optimize.minimize`.
     """
-    if not isinstance(nonlinear_inequality_constraints, list):
-        raise ValueError(
-            "`nonlinear_inequality_constraints` must be a list of callables, "
-            f"got {type(nonlinear_inequality_constraints)}."
-        )
 
     scipy_nonlinear_inequality_constraints = []
-    for nlc in nonlinear_inequality_constraints:
-        if _arrayify(nlc(x0)).item() < NLC_TOL:
+    for constraint in nonlinear_inequality_constraints:
+        if not isinstance(constraint, tuple):
+            raise ValueError(
+                f"A nonlinear constraint has to be a tuple, got {type(constraint)}."
+            )
+        if len(constraint) != 2:
+            raise ValueError(
+                "A nonlinear constraint has to be a tuple of length 2, "
+                f"got length {len(constraint)}."
+            )
+        nlc, is_intrapoint = constraint
+        if not nonlinear_constraint_is_feasible(
+            nlc, is_intrapoint=is_intrapoint, x=x0.reshape(shapeX)
+        ).all():
             raise ValueError(
                 "`batch_initial_conditions` must satisfy the non-linear inequality "
                 "constraints."
             )
-        f_obj, f_grad = _make_f_and_grad_nonlinear_inequality_constraints(
-            f_np_wrapper=f_np_wrapper, nlc=nlc
-        )
-        scipy_nonlinear_inequality_constraints.append(
-            {
-                "type": "ineq",
-                "fun": f_obj,
-                "jac": f_grad,
-            }
+
+        scipy_nonlinear_inequality_constraints += _make_nonlinear_constraints(
+            f_np_wrapper=f_np_wrapper,
+            nlc=nlc,
+            is_intrapoint=is_intrapoint,
+            shapeX=shapeX,
         )
     return scipy_nonlinear_inequality_constraints
+
+
+def evaluate_feasibility(
+    X: Tensor,
+    inequality_constraints: list[tuple[Tensor, Tensor, float]] | None = None,
+    equality_constraints: list[tuple[Tensor, Tensor, float]] | None = None,
+    nonlinear_inequality_constraints: list[tuple[Callable, bool]] | None = None,
+    tolerance: float = CONST_TOL,
+) -> Tensor:
+    r"""Evaluate feasibility of candidate points (within a tolerance).
+
+    Args:
+        X: The candidate tensor of shape `batch x q x d`.
+        inequality_constraints: A list of tuples (indices, coefficients, rhs),
+            with each tuple encoding an inequality constraint of the form
+            `\sum_i (X[indices[i]] * coefficients[i]) >= rhs`. `indices` and
+            `coefficients` should be torch tensors. See the docstring of
+            `make_scipy_linear_constraints` for an example. When q=1, or when
+            applying the same constraint to each candidate in the batch
+            (intra-point constraint), `indices` should be a 1-d tensor.
+            For inter-point constraints, in which the constraint is applied to the
+            whole batch of candidates, `indices` must be a 2-d tensor, where
+            in each row `indices[i] =(k_i, l_i)` the first index `k_i` corresponds
+            to the `k_i`-th element of the `q`-batch and the second index `l_i`
+            corresponds to the `l_i`-th feature of that element.
+        equality_constraints: A list of tuples (indices, coefficients, rhs),
+            with each tuple encoding an equality constraint of the form
+            `\sum_i (X[indices[i]] * coefficients[i]) = rhs`. See the docstring of
+            `make_scipy_linear_constraints` for an example.
+        nonlinear_inequality_constraints: A list of tuples representing the nonlinear
+            inequality constraints. The first element in the tuple is a callable
+            representing a constraint of the form `callable(x) >= 0`. In case of an
+            intra-point constraint, `callable()`takes in an one-dimensional tensor of
+            shape `d` and returns a scalar. In case of an inter-point constraint,
+            `callable()` takes a two dimensional tensor of shape `q x d` and again
+            returns a scalar. The second element is a boolean, indicating if it is an
+            intra-point or inter-point constraint (`True` for intra-point. `False` for
+            inter-point). For more information on intra-point vs inter-point
+            constraints, see the docstring of the `inequality_constraints` argument.
+        tolerance: The tolerance used to check the feasibility of constraints.
+            For inequality constraints, we check if `const(X) >= rhs - tolerance`.
+            For equality constraints, we check if `abs(const(X) - rhs) < tolerance`.
+            For non-linear inequality constraints, we check if `const(X) >= -tolerance`.
+            This avoids marking the candidates as infeasible due to tiny violations.
+
+    Returns:
+        A boolean tensor of shape `batch` indicating if the corresponding candidate of
+        shape `q x d` is feasible.
+    """
+    is_feasible = torch.ones(X.shape[:-2], device=X.device, dtype=torch.bool)
+    if inequality_constraints is not None:
+        for idx, coef, rhs in inequality_constraints:
+            if idx.ndim == 1:
+                # Intra-point constraints.
+                is_feasible &= (
+                    (X[..., idx] * coef).sum(dim=-1) >= rhs - tolerance
+                ).all(dim=-1)
+            else:
+                # Inter-point constraints.
+                is_feasible &= (X[..., idx[:, 0], idx[:, 1]] * coef).sum(
+                    dim=-1
+                ) >= rhs - tolerance
+    if equality_constraints is not None:
+        for idx, coef, rhs in equality_constraints:
+            if idx.ndim == 1:
+                # Intra-point constraints.
+                is_feasible &= (
+                    ((X[..., idx] * coef).sum(dim=-1) - rhs).abs() < tolerance
+                ).all(dim=-1)
+            else:
+                # Inter-point constraints.
+                is_feasible &= (
+                    (X[..., idx[:, 0], idx[:, 1]] * coef).sum(dim=-1) - rhs
+                ).abs() < tolerance
+    if nonlinear_inequality_constraints is not None:
+        for const, intra in nonlinear_inequality_constraints:
+            is_feasible &= nonlinear_constraint_is_feasible(
+                nonlinear_inequality_constraint=const,
+                is_intrapoint=intra,
+                x=X,
+                tolerance=tolerance,
+            )
+    return is_feasible
