@@ -10,7 +10,7 @@ Posterior module to be used with PyTorch distributions.
 
 from __future__ import annotations
 
-from typing import Any, Optional
+from typing import Any
 
 import torch
 from botorch.posteriors.posterior import Posterior
@@ -42,7 +42,7 @@ class TorchPosterior(Posterior):
 
     def rsample(
         self,
-        sample_shape: Optional[torch.Size] = None,
+        sample_shape: torch.Size | None = None,
     ) -> Tensor:
         r"""Sample from the posterior (with gradients).
 
@@ -78,6 +78,20 @@ class TorchPosterior(Posterior):
         """
         return getattr(self.distribution, name)
 
+    def __getstate__(self) -> dict[str, Any]:
+        r"""A minimal utility to support pickle protocol.
+
+        Pickle uses `__get/setstate__` to serialize / deserialize the objects.
+        Since we define `__getattr__` above, it takes precedence over these
+        methods, and we end up in an infinite loop unless we also define
+        `__getstate__` and `__setstate__`.
+        """
+        return self.__dict__
+
+    def __setstate__(self, d: dict[str, Any]) -> None:
+        r"""A minimal utility to support pickle protocol."""
+        self.__dict__ = d
+
     def quantile(self, value: Tensor) -> Tensor:
         r"""Compute quantiles of the distribution.
 
@@ -99,7 +113,8 @@ class TorchPosterior(Posterior):
         return self.log_prob(value).exp()
 
     def _extended_shape(
-        self, sample_shape: torch.Size = torch.Size()  # noqa: B008
+        self,
+        sample_shape: torch.Size = torch.Size(),  # noqa: B008
     ) -> torch.Size:
         r"""Returns the shape of the samples produced by the distribution with
         the given `sample_shape`.
