@@ -5,11 +5,8 @@
 # LICENSE file in the root directory of this source tree.
 
 
-from typing import Any, Type, Union
-
 import torch
 from botorch.logging import logger
-from botorch.posteriors.deterministic import DeterministicPosterior
 from botorch.posteriors.ensemble import EnsemblePosterior
 from botorch.posteriors.gpytorch import GPyTorchPosterior
 from botorch.posteriors.posterior import Posterior
@@ -17,7 +14,6 @@ from botorch.posteriors.posterior_list import PosteriorList
 from botorch.posteriors.torch import TorchPosterior
 from botorch.posteriors.transformed import TransformedPosterior
 from botorch.sampling.base import MCSampler
-from botorch.sampling.deterministic import DeterministicSampler
 from botorch.sampling.index_sampler import IndexSampler
 from botorch.sampling.list_sampler import ListSampler
 from botorch.sampling.normal import (
@@ -33,7 +29,7 @@ from torch.quasirandom import SobolEngine
 
 def _posterior_to_distribution_encoder(
     posterior: Posterior,
-) -> Union[Type[Distribution], Type[Posterior]]:
+) -> type[Distribution] | type[Posterior]:
     r"""An encoder returning the type of the distribution for `TorchPosterior`
     and the type of the posterior for the rest.
     """
@@ -46,7 +42,10 @@ GetSampler = Dispatcher("get_sampler", encoder=_posterior_to_distribution_encode
 
 
 def get_sampler(
-    posterior: TorchPosterior, sample_shape: torch.Size, **kwargs: Any
+    posterior: TorchPosterior,
+    sample_shape: torch.Size,
+    *,
+    seed: int | None = None,
 ) -> MCSampler:
     r"""Get the sampler for the given posterior.
 
@@ -58,24 +57,26 @@ def get_sampler(
         sample_shape: The sample shape of the samples produced by the
             given sampler. The full shape of the resulting samples is
             given by `posterior._extended_shape(sample_shape)`.
-        kwargs: Optional kwargs, passed down to the samplers during construction.
+        seed: Seed used to initialize sampler.
 
     Returns:
         The `MCSampler` object for the given posterior.
     """
-    kwargs["sample_shape"] = sample_shape
-    return GetSampler(posterior, **kwargs)
+    return GetSampler(posterior, sample_shape=sample_shape, seed=seed)
 
 
 @GetSampler.register(MultivariateNormal)
 def _get_sampler_mvn(
-    posterior: GPyTorchPosterior, sample_shape: torch.Size, **kwargs: Any
+    posterior: GPyTorchPosterior,
+    sample_shape: torch.Size,
+    *,
+    seed: int | None = None,
 ) -> NormalMCSampler:
     r"""The Sobol normal sampler for the `MultivariateNormal` posterior.
 
     If the output dim is too large, falls back to `IIDNormalSampler`.
     """
-    sampler = SobolQMCNormalSampler(sample_shape=sample_shape, **kwargs)
+    sampler = SobolQMCNormalSampler(sample_shape=sample_shape, seed=seed)
     collapsed_shape = sampler._get_collapsed_shape(posterior=posterior)
     base_collapsed_shape = collapsed_shape[len(sample_shape) :]
     if base_collapsed_shape.numel() > SobolEngine.MAXDIM:
@@ -83,51 +84,52 @@ def _get_sampler_mvn(
             f"Output dim {base_collapsed_shape.numel()} is too large for the "
             "Sobol engine. Using IIDNormalSampler instead."
         )
-        sampler = IIDNormalSampler(sample_shape=sample_shape, **kwargs)
+        sampler = IIDNormalSampler(sample_shape=sample_shape, seed=seed)
     return sampler
 
 
 @GetSampler.register(TransformedPosterior)
 def _get_sampler_derived(
-    posterior: TransformedPosterior, sample_shape: torch.Size, **kwargs: Any
+    posterior: TransformedPosterior,
+    sample_shape: torch.Size,
+    *,
+    seed: int | None = None,
 ) -> MCSampler:
     r"""Get the sampler for the underlying posterior."""
     return get_sampler(
-        posterior=posterior._posterior, sample_shape=sample_shape, **kwargs
+        posterior=posterior._posterior,
+        sample_shape=sample_shape,
+        seed=seed,
     )
 
 
 @GetSampler.register(PosteriorList)
 def _get_sampler_list(
-    posterior: PosteriorList, sample_shape: torch.Size, **kwargs: Any
+    posterior: PosteriorList, sample_shape: torch.Size, *, seed: int | None = None
 ) -> MCSampler:
     r"""Get the `ListSampler` with the appropriate list of samplers."""
     samplers = [
-        get_sampler(posterior=p, sample_shape=sample_shape, **kwargs)
+        get_sampler(posterior=p, sample_shape=sample_shape, seed=seed)
         for p in posterior.posteriors
     ]
     return ListSampler(*samplers)
 
 
-@GetSampler.register(DeterministicPosterior)
-def _get_sampler_deterministic(
-    posterior: DeterministicPosterior, sample_shape: torch.Size, **kwargs: Any
-) -> MCSampler:
-    r"""Get the dummy `DeterministicSampler` for the `DeterministicPosterior`."""
-    return DeterministicSampler(sample_shape=sample_shape, **kwargs)
-
-
 @GetSampler.register(EnsemblePosterior)
 def _get_sampler_ensemble(
-    posterior: EnsemblePosterior, sample_shape: torch.Size, **kwargs: Any
+    posterior: EnsemblePosterior,
+    sample_shape: torch.Size,
+    seed: int | None = None,
 ) -> MCSampler:
     r"""Get the `IndexSampler` for the `EnsemblePosterior`."""
-    return IndexSampler(sample_shape=sample_shape, **kwargs)
+    return IndexSampler(sample_shape=sample_shape, seed=seed)
 
 
 @GetSampler.register(object)
 def _not_found_error(
-    posterior: Posterior, sample_shape: torch.Size, **kwargs: Any
+    posterior: Posterior,
+    sample_shape: torch.Size,
+    seed: int | None = None,
 ) -> None:
     raise NotImplementedError(
         f"A registered `MCSampler` for posterior {posterior} is not found. You can "
