@@ -18,6 +18,7 @@ from botorch.exceptions import UnsupportedError
 from botorch.posteriors import Posterior
 from botorch.posteriors.higher_order import HigherOrderGPPosterior
 from botorch.posteriors.multitask import MultitaskGPPosterior
+from botorch.posteriors.transformed import TransformedPosterior
 from botorch.sampling.base import MCSampler
 from botorch.utils.sampling import draw_sobol_normal_samples, manual_seed
 from torch import Tensor
@@ -65,7 +66,7 @@ class NormalMCSampler(MCSampler, ABC):
         pass  # pragma: no cover
 
     def _update_base_samples(
-        self, posterior: Posterior, base_sampler: NormalMCSampler
+        self, posterior: Posterior, base_sampler: MCSampler
     ) -> None:
         r"""Update the sampler to use the original base samples for X_baseline.
 
@@ -101,7 +102,15 @@ class NormalMCSampler(MCSampler, ABC):
             expanded_samples = current_base_samples.view(view_shape).expand(
                 expanded_shape
             )
-            if isinstance(posterior, (HigherOrderGPPosterior, MultitaskGPPosterior)):
+            non_transformed_posterior = (
+                posterior._posterior
+                if isinstance(posterior, TransformedPosterior)
+                else posterior
+            )
+            if isinstance(
+                non_transformed_posterior,
+                (HigherOrderGPPosterior, MultitaskGPPosterior),
+            ):
                 n_train_samples = current_base_samples.shape[-1] // 2
                 # The train base samples.
                 self.base_samples[..., :n_train_samples] = expanded_samples[
@@ -112,17 +121,18 @@ class NormalMCSampler(MCSampler, ABC):
                     ..., -n_train_samples:
                 ]
             else:
+                batch_shape = non_transformed_posterior.batch_shape
                 single_output = (
-                    len(posterior.base_sample_shape) - len(posterior.batch_shape)
+                    len(posterior.base_sample_shape) - len(batch_shape)
                 ) == 1
                 if single_output:
-                    self.base_samples[
-                        ..., : current_base_samples.shape[-1]
-                    ] = expanded_samples
+                    self.base_samples[..., : current_base_samples.shape[-1]] = (
+                        expanded_samples
+                    )
                 else:
-                    self.base_samples[
-                        ..., : current_base_samples.shape[-2], :
-                    ] = expanded_samples
+                    self.base_samples[..., : current_base_samples.shape[-2], :] = (
+                        expanded_samples
+                    )
 
 
 class IIDNormalSampler(NormalMCSampler):
@@ -164,7 +174,7 @@ class SobolQMCNormalSampler(NormalMCSampler):
     r"""Sampler for quasi-MC N(0,1) base samples using Sobol sequences.
 
     Example:
-        >>> sampler = SobolQMCNormalSampler(1024, seed=1234)
+        >>> sampler = SobolQMCNormalSampler(torch.Size([1024]), seed=1234)
         >>> posterior = model.posterior(test_X)
         >>> samples = sampler(posterior)
     """
