@@ -7,7 +7,7 @@
 from __future__ import annotations
 
 from itertools import combinations
-from typing import Any, Optional
+from typing import Any
 
 import numpy as np
 import torch
@@ -81,7 +81,7 @@ class PairwiseIIDNormalSampler(PairwiseMCSampler, IIDNormalSampler):
     def __init__(
         self,
         sample_shape: torch.Size,
-        seed: Optional[int] = None,
+        seed: int | None = None,
         max_num_comparisons: int = None,
         **kwargs: Any,
     ) -> None:
@@ -103,7 +103,7 @@ class PairwiseSobolQMCNormalSampler(PairwiseMCSampler, SobolQMCNormalSampler):
     def __init__(
         self,
         sample_shape: torch.Size,
-        seed: Optional[int] = None,
+        seed: int | None = None,
         max_num_comparisons: int = None,
         **kwargs: Any,
     ) -> None:
