@@ -10,14 +10,12 @@ from botorch.utils.objective import apply_constraints, get_objective_weights_tra
 from botorch.utils.rounding import approximate_round
 from botorch.utils.sampling import (
     batched_multinomial,
-    construct_base_samples,
-    construct_base_samples_from_posterior,
     draw_sobol_normal_samples,
     draw_sobol_samples,
     manual_seed,
 )
 from botorch.utils.transforms import (
-    squeeze_last_dim,
+    average_over_ensemble_models,
     standardize,
     t_batch_mode_transform,
 )
@@ -26,16 +24,14 @@ from botorch.utils.transforms import (
 __all__ = [
     "apply_constraints",
     "approximate_round",
+    "average_over_ensemble_models",
     "batched_multinomial",
-    "construct_base_samples",
-    "construct_base_samples_from_posterior",
     "draw_sobol_normal_samples",
     "draw_sobol_samples",
     "estimate_feasible_volume",
     "get_objective_weights_transform",
     "get_outcome_constraint_transforms",
     "manual_seed",
-    "squeeze_last_dim",
     "standardize",
     "t_batch_mode_transform",
 ]
